
package mosaic.sim.kernel.continuous;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import mosaic.sim.kernel.*;
import mosaic.sim.kernel.control.*;
import mosaic.sim.kernel.simevent.*;

/**
* Class implementing a guard for a DoubleVariable variable crossing a value.
* <p>
* Implements the interface StateEventComponent.
* <b>Events:</b><p>
* <dl>
* <dt><code>zeroCrossingEvent</code></dt>
* <dd></dd>
* </dl>
*/

public abstract class StateEventGuard  extends AbstractModelComponent
   implements StateEventComponent
   {
   /** contains all StateEventListeners */
   protected ArrayList listeners = new ArrayList();

   public StateEventGuard()
   {
      this (null, "stateEventGuard");
   }

  /** StateEventGuard constructor with label*/
  public StateEventGuard(String iLabel) {
      this(null, iLabel);
  }

  /** StateEventGuard constructor with owner model and label*/
  public StateEventGuard(ModelComponent model, String iLabel) {
    super(iLabel);
    enabled = true;
    if (model != null)
          model.addModelComponent(this);
    setForeground(Color.red);
    setFont(new Font("SansSerif", Font.PLAIN, 10));
  }


// enabled -----------------------------------------------------------------------------------------

  /** flag is true if this StateEventComponent is enabled */
   protected boolean enabled;

   /**
   * Returns true if enabled
   */
   public boolean isEnabled()
      { return enabled; }

   /**
   * Sets enable
   */
   public void setEnabled(boolean en)
      { enabled = en; }


// tL ----------------------------------------------------------------------------------------------

   /**
   * Gets the time of last event
   */
	 public double getTL()
      { return 0.0; }

   /**
   * Sets the time of last event
   */
	 public void setTL(double newTime)
      { ; }

// -----------------------------------------------------------------------------------------------------------
// interface Comparable --------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------

// select order  ---------------------------------------------------------------------------------------------

  /** variable to hold the selectOrder for the timer which is used as a priority
  * for tie breaking between several timer with equal next event time.
  */
  protected int selectOrder = 0;

  /** Gets the select order of the timer */
  public int getSelectOrder() {
      return selectOrder;
  }

  /** Sets the select order of the timer */
  public void setSelectOrder(int newOrder) {
    int oldSelectOrder = selectOrder;
    selectOrder= newOrder;
  }


// event -------------------------------------------------------------------------------------------

	 /**
   * Checks if an event occured.
   */
   abstract public boolean eventOccured();

   /**
   * Trigger an event
   */
	   public void triggerEvent() {
      StateEvent stateEvent = new StateEvent(this);

      for(int loop=0;loop<listeners.size();loop++)
         ((StateEventListener)listeners.get(loop)).processSimEvent(stateEvent);
	    }

// add and remove listeners ------------------------------------------------------------------------

	 /**
   * Add SimEventListener
   */
   public void addSimEventListener(SimEventListener simEventListener) {
      if (!listeners.contains(simEventListener))
         listeners.add(simEventListener);
      }

	 /**
   * Remove SimEventListener
   */
   public void removeSimEventListener(SimEventListener simEventListener) {
      if (listeners.contains(simEventListener))
         listeners.remove(simEventListener);
      }

   /**
   * Add StateEventListener
   */
   public void addStateEventListener(StateEventListener stateEventListener) {
     if (!listeners.contains(stateEventListener))
         listeners.add(stateEventListener);
   }

	 /**
   * Remove StateEventListener
   */
   public void removeStateEventListener(StateEventListener stateEventListener) {
      if (listeners.contains(stateEventListener))
         listeners.remove(stateEventListener);
      }

// paint -----------------------------------------------------------------------------------------------------

  /* Paints the timer */
  public void paintModelComponent(Graphics g) {
      int h = getSize().height;
      int w = getSize().width;
      g.setColor(getForeground());
      g.drawOval(0, 0, w-1, h-1);
      g.drawLine(w/2, h/2, w-1, 0);
      g.setFont(getFont());
      FontMetrics fm = g.getFontMetrics();
      g.setColor(getForeground());
      String lbl = getLabel();
      if (lbl == null) {
          lbl = "V";
      }
      g.drawString(lbl,2,h-fm.getMaxDescent());
  }

//size: ------------------------------------------------------------------------------------------------------

  /** Gets the minimum size for the LayoutManager */
  public Dimension getMinimumSize() {
    return new Dimension(5,5);
  }

  /** Gets the preferred size for the LayoutManager */
  public Dimension getPreferredSize() {
    return new Dimension(16,12);
  }


//debug: -----------------------------------------------------------------------------------------------------

  /** Gives a textual description of the value of the variable as debug information. */
  protected String getDebugText() {
    return ("");
  }

}
