package mosaic.sim.kernel.control;

/**
 * All views that want to perform realtime Animation steps between the simulation
 * steps should implement this interface.
 * <p>
 * Those views should know that performAnimationStep will be called by
 * an ViewAnimatorThread for each frame.
 * In this Method the views should post reapintMessages if necessary.
 * if they posted repaints, they must notify the ViewAnimatorThread after the
 * paint() is done for the current frame.
 * Use a boolean to avoid duplicate notifications.
 *
 * This could be done like this:
 * <code>
 * public void boolean performAnimation(Time simTime) {
 *      boolean needRepaints = false;
 *      //do calculation and
 *      if (needRepaints) {
 *          curFramePainted = false;
 *          //call repaints here
 *      }
 *  return !curFramePainted;
 *  }
 *
 *
 * public void paint(Graphics g) {
 *      // do paint
 *      if (!curFramePainted) {
 *          animator.nextViewPainted();
 *      }
 * }
 *
 * boolean synchronized curFramePainted = true;
 * ViewAnimatorThread animator;
 * </code>
 *
 * @author Thomas Hilpold
 */
public interface AnimatedView {

    /**
    * Performs necessary calculation and posts repaint messages to the
    * AWTEventThread depending on the given time.
    * <p>
    * Will be called by an AnimatorThread for each frame to be painted
    * @param time current simulation-time
    * @return false, if no repaint was called
    */
    public boolean performAnimationStep(double time);

    public void setPerformAnimation(boolean animate);

    public boolean getPerformAnimation();
}
