package mosaic.sim.kernel.control;



import mosaic.util.*;

public class RealTimeStrategy extends SchedulingStrategy {

    public static boolean[] DEBUG = new boolean[] {true,
                                                false,
                                                true};

    RealTimeStrategy(SimulationScheduler scheduler) {
        this.scheduler = scheduler;
        targetFps = 35;
        regulationFactor = 1.1;
        reset();
    }

    //
    // Animation
    //

    private double avgAnimationDuration;    //millis
    private double lastAnimationSimTime;    //secs
    private double targetFps;               // Frames per second between steps
    private double currentFps;              // Frames per second between steps
    private long frameDuration;
    private int nrOfAnimations;
    private long totalAnimationDuration;
    private long slidingAvgCounter;

    private double regulationFactor;

    //
    // User Properties
    //
    public double getTargetFps() {
        return targetFps;
    }

    public void setTargetFps(double targetFps) {
        if (targetFps > 0.0) {
            this.targetFps = targetFps;
            frameDuration = (long) (1000.0 / targetFps);
        }
    }

    public double getRegulationFactor() {
        return regulationFactor;
    }

    public void setRegulationFactor(double reg) {
        if (reg > 1.0) {
            this.regulationFactor = reg;
        }
    }

    //
    // Simulation
    //
    void reset() {
        avgAnimationDuration = 10;
        lastAnimationSimTime = 0;
        currentFps = 20;
        frameDuration = (long) (1000.0 / targetFps);
        nrOfAnimations = 0;
        totalAnimationDuration = 0;
        slidingAvgCounter = 0;
    }

    public void performScheduling() {
        Thread.currentThread().setPriority(Thread.NORM_PRIORITY);
        double t2 = scheduler.getSimApplication().getSimulator().getNextSimulationTime();
        double simRealTimeSecs;

        /* run until scheduler says stop */
        do {
            scheduler.updateCurrentAnimationRate();
            simRealTimeSecs = scheduler.getSimRealTime();

            /* animate until the event time is reached */
            animateUntilSimTime(t2);
            scheduler.isLate = ((simRealTimeSecs - t2) > 0.1);
            if  (scheduler.isLate){
                System.out.println(" getSimApplication().getSimulator() too slow dev:"
                     + (simRealTimeSecs - t2));
            }

            /* trigger the getSimApplication().getSimulator() to perform one step */
            scheduler.getSimApplication().getSimulator().step();
            scheduler.analyzer.newStep();

            /* get next Time */
            t2 = scheduler.getSimApplication().getSimulator().getNextSimulationTime();
        } while(!scheduler.stopOnNextOccasion);
    }

    //
    // Animation
    //

    private void animateUntilSimTime(double stopSimTime) {
        long stopRealTime;
        long animationDuration;
        long sleepDuration;
        long animateStartTime;
        long preSleep;
        long postSleep;

        animationDuration = (long)avgAnimationDuration;
        stopRealTime = System.currentTimeMillis() +
                            scheduler.toRealTimeDuration(stopSimTime);


        /* animate until time simTime happens */
        LogPrinter.println(3, "****** ANIM START *******");
        while (System.currentTimeMillis() < stopRealTime -
                frameDuration) {

            //
            // Immediately after STEP() sleep until animation schould start
            //
            preSleep = (long) ((frameDuration - avgAnimationDuration) / 2.0);
            try {
                Thread.currentThread().sleep(preSleep);
            } catch (InterruptedException e) {
            }
            animateStartTime = System.currentTimeMillis();

            // perform Animation
            scheduler.animator.animate(scheduler.getSimRealTime());
            animationDuration = System.currentTimeMillis() - animateStartTime;
            nrOfAnimations++;
            totalAnimationDuration += animationDuration;
            slidingAvgCounter++;
            avgAnimationDuration = totalAnimationDuration / (double) slidingAvgCounter;
            if (slidingAvgCounter > 10) {
                totalAnimationDuration = animationDuration;
                slidingAvgCounter = 1;
            }
            if (DEBUG[1]) System.out.print(" an = " + animationDuration + "ms");

            // calculate how long the thread can go to sleep in consideration of current fps
            sleepDuration = frameDuration - animationDuration;
            // add data concerning this frame to the Analyzer
            scheduler.analyzer.newFrame();
            scheduler.analyzer.addAnimationDuration(animationDuration);
            scheduler.analyzer.addSleepDuration(sleepDuration);
            scheduler.analyzer.addFrameDuration(frameDuration);
            if (sleepDuration < 0) {

                // DECREASE FPS to 90 %
                currentFps = currentFps * (1 / regulationFactor);
                sleepDuration = 0;
                frameDuration = (long) (1000.0 / currentFps);
                if (DEBUG[2]) System.err.println(" dec FPS to " + (int) currentFps);
            }
            else if ((sleepDuration > 20) && (currentFps < targetFps)) {

                // INCREASE FPS to 115 % if we have not reached targetFps yet
                // (accuracy not necessary)
                currentFps = currentFps * regulationFactor;
                frameDuration = (long) (1000.0 / currentFps);
                if (DEBUG[2]) {
                    System.err.println(" inc FPS to " + (int) currentFps);
                }
            }
            else {

                // KEEP CURRENT FPS
            }
            postSleep = sleepDuration - preSleep;
            if (postSleep < 0 ) postSleep = 0;
            try {
                // Sleep as long as we can afford in consideration of current FPS
                //sleepDuration - preSleep
                Thread.sleep(sleepDuration - preSleep);
            } catch (InterruptedException e) {
            }
            LogPrinter.println(3, "pre = " + preSleep + " ani = " + animationDuration +
                                  " post = " + postSleep + " frm = " + frameDuration);

        } //while
        LogPrinter.println(3, "****** ANIM STOP *******");
    }
}



