package mosaic.sim.kernel.control;

import mosaic.util.*;

public class RealTimeStrategy3 extends SchedulingStrategy {

    final static double TOO_SLOW_TOLERANCE = 0.5;       // simtime

    public static boolean[] DEBUG = new boolean[] {
                                                false,
                                                false,
                                                false,  // out thread info
                                                false,  // create LOG Files
                                                true    // trace animationrealtime
                                                };

    //
    // Data + Higheperformancecounter
    //

    private double totalSleepPerStep;
    private double targetFps;              // Frames per second between steps
    private double currentFps;      // Frames per second between steps
    private double frameDuration;
    private int nrOfAnimations;
    private double totalAnimationDuration;
    private double slidingAvgCounter;
    private double avgFrameDuration;
    private HighPerformanceCounter hpCounter;

    //
    // Construction
    //

    RealTimeStrategy3(SimulationScheduler scheduler) {
        this.scheduler = scheduler;
        hpCounter = new HighPerformanceCounter();
        targetFps = 35;     // Frames per second between steps
        reset();
    }

    //
    // User Properties
    //
    public double getTargetFps() {
        return targetFps;
    }

    public void setTargetFps(double targetFps) {
        if (targetFps > 0.0) {
            this.targetFps = targetFps;
        }
    }

    //
    // Simulation
    //

    void reset() {
        totalSleepPerStep = 0;
        currentFps = 20;                     // Frames per second between steps
        frameDuration = (1000.0 / targetFps);//millisecs
        nrOfAnimations = 0;
        totalAnimationDuration = 0;
        slidingAvgCounter = 0;
        avgFrameDuration = frameDuration;
    }

    void performScheduling() {
        if (DEBUG[2]) SimInformer.printAllThreads();
        Thread.currentThread().setPriority(Thread.NORM_PRIORITY);
        double t2 = scheduler.getSimApplication().getSimulator().getNextSimulationTime();
        double simRealTimeSecs;

        ///* set curSimTime to StartTime */

        /* run until scheduler says stop */
        do {
            scheduler.updateCurrentAnimationRate();
            simRealTimeSecs = scheduler.getSimRealTime();
            totalSleepPerStep = 0.0;

            if (DEBUG[3]) LogPrinter.println(4, "- simRealTime = " + ft(simRealTimeSecs));
            /* animate until the event time is reached */
            animateUntilSimTime(t2);
            if (DEBUG[3]) LogPrinter.println(4, "- T2          = " + ft(t2));
            scheduler.isLate = ((simRealTimeSecs - t2) > TOO_SLOW_TOLERANCE);
            if  (scheduler.isLate){
                LogPrinter.println(" Sim slow dev:"
                     + (simRealTimeSecs - t2));
            }
            scheduler.analyzer.newStep();
            scheduler.analyzer.addSleepDuration(totalSleepPerStep);
            hpCounter.resetTime();
            /* trigger the getSimApplication().getSimulator() to perform one step */
            scheduler.getSimApplication().getSimulator().step();
            scheduler.analyzer.addStepDuration(hpCounter.getTime());
            /* get next Time */
            t2 = scheduler.getSimApplication().getSimulator().getNextSimulationTime();
        } while(!scheduler.stopOnNextOccasion);
    }

    //
    // Animation
    //

    private void animateUntilSimTime(double stopSimTime) {
        long stopRealTime;
        double animationDuration;
        double sleepDuration;
        long animateStartTime;
        double nextAnimTime;
        double realFrameDuration;
        double curSimTime;

        stopRealTime = System.currentTimeMillis() +
                            scheduler.toRealTimeDuration(stopSimTime);

        if (DEBUG[4]) System.out.println("animationRT for " + (stopRealTime - System.currentTimeMillis()));
        /* animate until time simTime happens */
        if (DEBUG[3]) LogPrinter.println(3, "****** ANIM START *******");
        curSimTime = 0; //only for first comparison
        while (System.currentTimeMillis() < (stopRealTime -
                avgFrameDuration) && (curSimTime < stopSimTime)) {

            //
            // Immediately after STEP() sleep until animation schould start
            //
            animateStartTime = System.currentTimeMillis();
            hpCounter.resetTime();
            // perform Animation
            curSimTime = scheduler.getSimRealTime();
            if (DEBUG[3]) LogPrinter.println(4, "Ax = " + ft(curSimTime));
            scheduler.animator.animate(curSimTime);
            animationDuration = hpCounter.getTime();
            sleepDuration = frameDuration - animationDuration;
            if (sleepDuration <= 0) {
                sleepDuration = 1;
                if (DEBUG[3]) LogPrinter.print(3, " **** SLEEP  <= 0 ****");
            }
            nextAnimTime = animateStartTime + frameDuration;
            try {
                Thread.currentThread().sleep(Math.round(sleepDuration));
                totalSleepPerStep += sleepDuration;
            } catch (InterruptedException e) {
            }

            /* AWT finished - determine time deviation  */
            realFrameDuration = System.currentTimeMillis() - animateStartTime;
            if (realFrameDuration > avgFrameDuration) {
                avgFrameDuration += (realFrameDuration - avgFrameDuration) * 0.025;
                if (DEBUG[3]) LogPrinter.print(3, " new MAX = "  + avgFrameDuration);
            } else {
                avgFrameDuration -= (avgFrameDuration - realFrameDuration) * 0.025;
            }
            if (DEBUG[3]) LogPrinter.print(3, " Dev = "  + (frameDuration - realFrameDuration));

            nrOfAnimations++;
            totalAnimationDuration += animationDuration;
            if (DEBUG[1]) System.out.print(" an = " + animationDuration + "ms");

            // add data concerning this frame to the Analyzer
            scheduler.analyzer.newFrame();
            scheduler.analyzer.addAnimationDuration(animationDuration);
            scheduler.analyzer.addFrameDuration(frameDuration);
        } //while
        if (DEBUG[3]) LogPrinter.println(3, "****** ANIM STOP *******");
        long postSleep = stopRealTime - System.currentTimeMillis();
        if (postSleep > 0) {
            try {
                Thread.sleep(postSleep);
                totalSleepPerStep += postSleep;
                if (DEBUG[3]) LogPrinter.println(4, "POST = " + postSleep);
            } catch (InterruptedException e){
            }
        }
    }

    private String ft(double in) {
        return ("" + ((int) (in * 1000)));
    }
}



