package mosaic.sim.kernel.control;

import sim.util.*;
import mosaic.sim.util.thread.ThreadEventDispatcher;

/**
 * RealtimeStrategy with external event support
 * (for Agent package)
 *
 * changes for integratino of external events appear in lines:
 * 84 and 125
 * Hilpold 7.7.2000
 */
public class RealTimeStrategy3_Agent extends SchedulingStrategy {

    final static double TOO_SLOW_TOLERANCE = 0.5;       // simtime

    //
    // Data + Highperformancecounter
    //

    private double targetFps;              // Frames per second between steps
    private double currentFps;      // Frames per second between steps
    private double frameDuration;
    private double avgFrameDuration;
    private HighPerformanceCounter hpCounter;

    //
    // Construction
    //

    RealTimeStrategy3_Agent(SimulationScheduler scheduler) {
        this.scheduler = scheduler;
        hpCounter = new HighPerformanceCounter();
        targetFps = 35.0;     // Frames per second between steps
        reset();
    }

    //
    // User Properties
    //

    public double getTargetFps() {
        return targetFps;
    }

    public void setTargetFps(double targetFps) {
        if (targetFps > 0.0) {
            this.targetFps = targetFps;
        }
    }

    //
    // Simulation
    //

    void reset() {
        currentFps = 20;                      // Frames per second between steps
        frameDuration = (1000.0 / targetFps); //millisecs
        avgFrameDuration = frameDuration;
    }

    void performScheduling() {
        Thread.currentThread().setPriority(Thread.NORM_PRIORITY);
        double t2 = scheduler.getSimApplication().getSimulator().getNextSimulationTime();
        double simRealTimeSecs;
        ThreadEventDispatcher ted=scheduler.getSimApplication().getSimulator().getThreadEventDispatcher();
        Simulator sim=scheduler.getSimApplication().getSimulator();

        ///* set curSimTime to StartTime */

        /* run until scheduler says stop */
        do {
            scheduler.updateCurrentAnimationRate();
            simRealTimeSecs = scheduler.getSimRealTime();

            /*
             * animate until the event time is reached or external events are available
             * and need to be queued
             */
            animateUntilSimTime(t2);
            /**
             * CHANGE for external events
             */
            while (!ted.isEmpty()) {

//                agent.processEventsAt(scheduler.getSimRealTime());
                scheduler.getSimApplication().setTime(scheduler.getSimRealTime());
                ted.processAll();
                t2 = sim.getNextSimulationTime();
                animateUntilSimTime(t2);
            }

            scheduler.isLate = ((simRealTimeSecs - t2) > TOO_SLOW_TOLERANCE);
            hpCounter.resetTime();
            /* trigger the sim.Simulation.getSimulator() to perform one step */
            scheduler.getSimApplication().getSimulator().step();
            /* get next Time */
            t2 = scheduler.getSimApplication().getSimulator().getNextSimulationTime();
        } while(!scheduler.stopOnNextOccasion);
    }

    //
    // Animation
    //

    private void animateUntilSimTime(double stopSimTime) {
        long stopRealTime;          // time when the animation should stop
        double animationDuration;   // the duration the animator needs to paint one frame
        double sleepDuration;       // the duration the thread schould sleep after paint
        long animateStartTime;      // the time before the animator starts painting this frame
        double nextAnimTime;        // the nest time the animator schould paint
        double realFrameDuration;   // the real duration of the whole frame
        double curSimTime;          // the simRealTime of the simulator
        double breakForAgentEvents; // if true, the animation will stop immediatly, to process
                                    // external event

        stopRealTime = System.currentTimeMillis() +
                            scheduler.toRealTimeDuration(stopSimTime);

        /* animate until time simTime happens */
        curSimTime = 0; //only for first comparison
        ThreadEventDispatcher ted=scheduler.getSimApplication().getSimulator().getThreadEventDispatcher();
        while (System.currentTimeMillis() < (stopRealTime -
                avgFrameDuration)
                && (curSimTime < stopSimTime) &&
                (ted.isEmpty())
                ) {

            //
            // Immediately after STEP() sleep until animation schould start
            //
            animateStartTime = System.currentTimeMillis();
            hpCounter.resetTime();
            // perform Animation
            curSimTime = scheduler.getSimRealTime();
            scheduler.animator.animate(curSimTime);
            animationDuration = hpCounter.getTime();
            sleepDuration = frameDuration - animationDuration;
            if (sleepDuration <= 0) {
                sleepDuration = 1;
            }
            nextAnimTime = animateStartTime + frameDuration;
            try {
                Thread current=Thread.currentThread();
                synchronized(current){
                    current.wait(Math.round(sleepDuration));
                }
            } catch (InterruptedException e) {
            }

            /* AWT finished - determine time deviation  */
            realFrameDuration = System.currentTimeMillis() - animateStartTime;
            if (realFrameDuration > avgFrameDuration) {
                avgFrameDuration += (realFrameDuration - avgFrameDuration) * 0.025;
            } else {
                avgFrameDuration -= (avgFrameDuration - realFrameDuration) * 0.025;
            }


        } //while
        long postSleep = stopRealTime - System.currentTimeMillis();
        if (postSleep > 0 && ted.isEmpty()) {
            try {
                Thread current=Thread.currentThread();
                synchronized(current){
                    current.wait(postSleep);
                }
            } catch (InterruptedException e){
            }
        }
    }

    private String ft(double in) {
        return ("" + ((int) (in * 1000)));
    }
}



