package mosaic.sim.kernel.control;

import java.util.*;
import mosaic.sim.kernel.StaticProjectSettings;
import mosaic.sim.*;
import mosaic.sim.kernel.SimApplication;
 /*
  * The SimulationsScheduler is responsible for the SimulationStrategy (mode of
  * simulation) and therefore tells the simulatior when to perform a
  * simulationstep and the registered AnimatedViews when to draw.
  * Animation and simulation is performed in just one Thread.
  */
public class SimulationScheduler extends Object implements Runnable {

    public static boolean[] DEBUG = new boolean[] {
                                                false
                                                };
    public static final int MILLISEC = 1;
    public static final int SEC = 1000;
    public static final int MIN = SEC * 60;
    public static final int HOUR = MIN * 60;
    public static final int DAY = HOUR * 24;
    public static final int WEEK = DAY * 7;

    private Thread schedulerThread;

    /** a strategy for scheduling */
    SchedulingStrategy strategy;
    String strategyAsText;
    Animator animator;
    volatile boolean isRunning;
    volatile boolean pauseOnNextOccasion;
    volatile boolean stopOnNextOccasion;
    volatile boolean isLate;

    //
    // Timing Data
    //
    volatile long lastRealTime;      // last change time in ms
    volatile long startRealTime;
    volatile double curSimTime;  // the incrementing current simulation Time
    volatile double simStartTime;  // the incrementing current simulation Time
    volatile double simStopTime;  // the incrementing current simulation Time

    volatile int timeUnit;  // millisec, sec, ...
    volatile double curAnimationRate;  // animation rate from the Simulator

    String timeUnitAsText;

    static SchedulingAnalyzer analyzer;

    static SchedulingAnalyzer getAnalyzer() {
        return analyzer;
    }

    //
    // Constructor
    //

    public SimulationScheduler(SimApplication application) {
        setSimApplication(application);
        strategy = null;
        animator = getSimApplication().getAnimator();
        isRunning = false;
        stopOnNextOccasion = false;
        pauseOnNextOccasion = false;
        analyzer = new SchedulingAnalyzer();
        if (DEBUG[0]) System.out.println(" SIMSCED CREATED ");
        setStrategyAsText(StaticProjectSettings.getSchedulingModes()[2]);
        setTimeUnitAsText(StaticProjectSettings.getTimeUnits()[1]);
    }

// SimApplication ----------------------------------------------------------------------------------

    protected SimApplication application;

    /** Gets the simulation application object */
    public SimApplication getSimApplication() {
        return application;
    }

    /** Sets the simulation application object */
    public  void setSimApplication(SimApplication application) {
        this.application = application;
    }

//animation rate -----------------------------------------------------------------------------------

    /**
     * Lock needed because long value handling is not thread safe on a 32 bit
     * machine.
     */
    Integer ANIMLOCK = new Integer(-1);

    /** variable to hold the time delay between events. Used for animation. */
    long animationRate= 10;

    /** Gets the time delay between events. */
    public long getAnimationRate() {
        synchronized(ANIMLOCK) {
            return animationRate;
        }
    }

    /** Gets the time delay between events. */
    public void setAnimationRate(long newAnimationRate) {
        synchronized(ANIMLOCK) {
            animationRate = newAnimationRate;
        }
    }


// scheduling strategy -----------------------------------------------------------------------------

    public void setStrategyAsText(String newStrat) {
        String[] modes = StaticProjectSettings.getSchedulingModes();
        boolean valid = false;
        if (!isRunning) {
            if (newStrat.equals(modes[0])) {
                strategy = new RealTimeStrategy3(this);
                valid = true;
            } else if (newStrat.equals(modes[1])) {
                strategy = new RealTimeStrategy(this);
                valid = true;
            } else if (newStrat.equals(modes[2])) {
                strategy = new RealTimeStrategyNoHPC(this);
                valid = true;
            } else if (newStrat.equals(modes[3])) {
                strategy = new FixedTimeStrategy(this);
                valid = true;
            } else if (newStrat.equals(modes[4])) {
                strategy = new FixedTimeStrategyNoHPC(this);
                valid = true;
            } else if (newStrat.equals(modes[5])) {
                strategy = new RealTimeStrategy3_Agent(this);
                valid = true;
            } else {
                System.err.println("*  MODE NOT RECOGNIZED *");
            }
            if (valid) strategyAsText = newStrat;
        } else {
            System.err.println("Cannot set scheduling mode while running. ");
        }
    }

    public SchedulingStrategy getSchedulingStrategy() {
        return strategy;
    }

    public String getStrategyAsText() {
        return strategyAsText;
    }

    public void setTimeUnitAsText(String tu) {
        String[] timeUnits = StaticProjectSettings.getTimeUnits();
        boolean valid = false;
        if (!isRunning) {
            if (tu.equals(timeUnits[0])) {
                timeUnit = MILLISEC;
                valid = true;
            } else if (tu.equals(timeUnits[1])) {
                timeUnit = SEC;
                valid = true;
            } else if (tu.equals(timeUnits[2])) {
                timeUnit = MIN;
                valid = true;
            } else if (tu.equals(timeUnits[3])) {
                timeUnit = HOUR;
                valid = true;
            } else if (tu.equals(timeUnits[4])) {
                timeUnit = DAY;
                valid = true;
            } else if (tu.equals(timeUnits[5])) {
                timeUnit = WEEK;
                valid = true;
            } else {
                System.err.println(" *  MODE NOT RECOGNIZED *");
            }
            if (valid) timeUnitAsText = tu;
        } else {
            System.err.println("Cannot set scheduling mode while running. ");
        }
    }

    public String getTimeUnitAsText() {
        return timeUnitAsText;
    }



    public void reset() {
        if (!isRunning()) {
            strategy.reset();
            curSimTime = simStartTime;
        }
    }

    public void start() {
        start(false);
    }

    private void start(boolean onlyOneStep) {
        /*
         * initialize sim-times only if called the first time
         * otherwise continue sim/anim from the last stop
         */
        strategy.reset();
        schedulerThread = new Thread (this, "Simulation-scheduler");
        schedulerThread.setPriority(Thread.NORM_PRIORITY);
        getSimApplication().getSimulator().setRunning(true);
        stopOnNextOccasion = onlyOneStep;
        schedulerThread.start();
        if (DEBUG[0]) System.err.println(" SCHEDULER STARTED ");
    }

    public void stop() {
        stopOnNextOccasion();
        while (isRunning()) {
            try {
                Thread.currentThread().sleep(100);
            } catch (InterruptedException e) {
            }
        }
        getSimApplication().getSimulator().setRunning(false);
        if (DEBUG[0]) System.err.println(" SCHEDULER STOPPED ");
    }

    public void step() {
        if (!isRunning()) {
             start(true);
             stop();
        }
    }

    public boolean isRunning() {
        return isRunning;
    }


    private void pauseOnNextOccasion() {
        pauseOnNextOccasion = true;

    }

    private void stopOnNextOccasion(){
        synchronized(this) {
            stopOnNextOccasion = true;
        }
    }

    private boolean isSchedulerRunning() {
        return isRunning;
    }

    /** schedules */
    public void run() {
        lastRealTime = System.currentTimeMillis();
        startRealTime = lastRealTime;
        isRunning = true;
        // stopOnNextOccation has to be true if only one step should be made
        strategy.performScheduling();
        isRunning = false;
    }

    /**
     * returns the animationRate
     * (called only by Schedulerthread)
     */
    public double getCurrentAnimationRate() {
        return curAnimationRate;
    }

    /**
     * gets the AnimationRate from the Simulator
     * the access will be thread safe (synced)
     * this should be called by the strategy
     * (called only by Schedulerthread)
     */
    public void updateCurrentAnimationRate() {
        curAnimationRate = getAnimationRate();
    }

    /**
     * calculates the simulation time in dependency of
     * 1.the real time
     * 2.the current animation rate
     */
    public double getSimRealTime() {
        long curTime;

        curTime = System.currentTimeMillis();
        if (curTime != lastRealTime) {
            curSimTime = curSimTime + (curTime - lastRealTime) *
                    curAnimationRate / 1000.0 / timeUnit;
            lastRealTime = curTime;
        }
        return  curSimTime;
    }


    /**
     * calculates the REAL time when the  given simulation time will be reached
     * in dependency of
     * 1.the given real time
     * 2.the current animation rate
     * Mark that any change in the animationrate invalidates previously calculated
     * realtimes.
     */
    protected long toRealTimeDuration(double untilSimTime) {
        double now = getSimRealTime();
        return Math.round(((untilSimTime - now) / curAnimationRate * 1000.0 *
                timeUnit));
    }

    public void timeDebug() {
        System.out.println("");
    }

    /**----------------------------------------------------------------------
    *
    * Analyzer
    *
    *----------------------------------------------------------------------*/
    protected class SchedulingAnalyzer extends Object {

        public SchedulingAnalyzer() {
            analyzerIntervall = 50;
        }

        void addSleepDuration(double lastDuration) {
            intervallSleepDuration += lastDuration;
        }

        void addFrameDuration(double lastDuration) {
            intervallFrameDuration += lastDuration;
        }

        void addAnimationDuration(double lastDuration) {
            intervallAnimationDuration += lastDuration;
        }

        void addStepDuration(double lastDuration) {
            intervallStepDuration += lastDuration;
        }

        void newStep() {
            if (stepCounter == analyzerIntervall) {
                stepsPerSecond = 1000 * ((double)analyzerIntervall /
                    (System.currentTimeMillis() - countStartRealTime));
                stepCounter = 0;
            }
            if (stepCounter == 0) {
                countStartRealTime = System.currentTimeMillis();
            }
            stepCounter++;
        }

        void newFrame() {
            totalFrames++;
            frameInIntervallCounter++;
            if (frameInIntervallCounter > analyzerIntervall) {
                frameInIntervallCounter = 1;
                intervallSleepDuration = 0;
                intervallAnimationDuration = 0;
                intervallFrameDuration = 0;
                intervallStepDuration = 0;
            }
        }

        public long getTotalFrames() {
            return totalFrames;
        }

        public double getAvgFps() {
            return 1000 / (intervallFrameDuration / frameInIntervallCounter);
        }

        public double getAvgSleepDuration() {
            return intervallSleepDuration / frameInIntervallCounter;
        }

        public double getAvgAnimationDuration() {
            return intervallAnimationDuration / frameInIntervallCounter;
        }

        public double getAvgFrameDuration() {
            return intervallFrameDuration / frameInIntervallCounter;
        }

        public double getAvgStepDuration() {
            return intervallStepDuration / frameInIntervallCounter;
        }

        public double getStepsPerSecond() {
            return  stepsPerSecond;
        }

        //
        // Analyzing data
        //
        protected volatile long analyzerIntervall; // in Frames
        protected transient volatile double intervallSleepDuration;
        protected transient volatile double intervallAnimationDuration;
        protected transient volatile double intervallFrameDuration;
        protected transient volatile double intervallStepDuration;
        protected transient volatile long frameInIntervallCounter;
        protected transient volatile long totalFrames;
        protected transient volatile long stepCounter;
        protected transient volatile long countStartRealTime;
        protected transient volatile double stepsPerSecond;
    }
}
