/**
 * Project mosaic
 *
 * Simulator.java
 *
 * copyright 2000, JKU Linz, Austria
 */

package mosaic.sim.kernel.control;

import mosaic.sim.kernel.*;
import mosaic.sim.kernel.variable.*;
import java.beans.*;
import java.io.*;
import java.util.*;
import mosaic.sim.kernel.output.VariableView;
import mosaic.sim.kernel.continuous.*;
import mosaic.sim.*;
import mosaic.sim.kernel.simevent.*;
import mosaic.sim.util.thread.*;

/**
 * Class realizing the simulator.
 * Contains the simulation steps with call to integration and event handling.
 * One singleton object is created in a simulation application.
 * <p>
 * Listens to model component events of all model components in the simulation application
 * to cope with the structural changes within the simulation. In particular, it maintains
 * internal information structures to be able to carry out the simulation (efficiently)
 * and forwards structural changes to the integrator, animator, etc. to allow them to do the same.
 */
public final class Simulator implements ModelComponentEventListener {


// Constructors ------------------------------------------------------------------------------------

    /** Constructor with simApplication as parameter */
    public Simulator(SimApplication application) {
        setSimApplication(application);
        nrEvents = new IntVariable(null, 0);
        handleAdditionOfModelComponent(null,getSimApplication().getRootModel());
///*af
        ted.setNofiyObject(Thread.currentThread());
//af*/
    }


// SimApplication ----------------------------------------------------------------------------------

    protected SimApplication application;

    /** Gets the simulation application object */
    public SimApplication getSimApplication() {
        return application;
    }

    /** Sets the simulation application object */
    public  void setSimApplication(SimApplication application) {
        this.application = application;
    }

//reset --------------------------------------------------------------------------------------------

    /**
    * Resets the simulator:
    * removes the current events
    * resets the simulation time,
    * resets the nr of events,
    */
    public void reset() {
        removeAllEvents();
        removeAllInputEvents();
        removeAllChangeEvents();
        resetNrEvents();
    }

// ThreadEventDispatcher ---------------------------------------------------------------------------

    /** The ThreadEventDispatcher is used to establish a source<->listener communication between
     *  components that operate in different threads. */
///*af
     protected ThreadEventDispatcher ted=new BasicThreadEventDispatcher();
//af*/

    /** Gets the ThreadEventDispatcher wich is used to establish a source<->listener communication between
     *  components that operate in different threads. */
///*af
    public ThreadEventDispatcher getThreadEventDispatcher(){
        return ted;
    }
//af*/


// -------------------------------------------------------------------------------------------------
// ------------ simulation -------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------

// simulation running ------------------------------------------------------------------------------

    protected boolean stateEventThrown;

    protected boolean eventThrown;


    transient boolean running= false;
    transient boolean suspended= false;

    /** Tests if running flag is set. */
    public boolean isRunning() {
        return running;
    }

    /** Sets the running flag. */
    void setRunning(boolean r) {
        running= r;
    }

    /** Tests if suspended flag is set. */
    public boolean isSuspended() {
        return suspended;
    }

    /** Sets the supended flag. */
    private void setSuspended(boolean s) {
        suspended= s;
    }


//event counter ------------------------------------------------------------------------------------

    /** variable to hold the nr of events in the current simulation run. */
    IntVariable nrEvents;

    /** Gets the nr of events in the current simulation run. */
    public int getNrEvents() {
        return nrEvents.getValue();
    }

    /** Sets the nr of events in the current simulation run. */
    public void setNrEvents(int newNrEvents) {
        nrEvents.setValue(newNrEvents);
    }

    /** Resets the nr of events in the current simulation run to 0. */
    void resetNrEvents() {
        setNrEvents(0);
    }

    /** Increments the nr of events in the current simulation run by 1. */
    void incNrEvents() {
        setNrEvents(nrEvents.getValue()+1);
    }

    /** Add a variable change listener to the nrEvents variable */
    public void addNrEventsListener(VariableChangeListener l) {
        nrEvents.addVariableChangeListener(l);
    }

    /** Remvoe a variable change listener to the nrEvents variable */
    public void removeNrEventsListener(VariableChangeListener l) {
        nrEvents.removeVariableChangeListener(l);
    }


// simulate step -----------------------------------------------------------------------------------

    /** makes one simulation step */
    final void step()  {

        synchronized (getSimApplication().stepLock) {
            getSimApplication().debugOutput("simulation", " start step at " +
                    getSimApplication().getTime() + "-----------------------------------", 2);

            eventThrown = false;

            // trigger all pending discrete variable changes first
            if (! variableChanges.isEmpty()) {
                getSimApplication().debugOutput("simulation", " !!!! executing pending variable changes !!!!! ", 3);
                fireVariableChanges();
                eventThrown = true;
            }

            // trigger all inputEvents - check: later triggered as ted (see below)
            while (!inputEventList.isEmpty()) {
                triggerNextInputEvent();
                eventThrown = true;
            }

            // trigger all events from other threads
///*af
            if (!ted.isEmpty())
                ted.processAll();
//af*/
            Integrator i = getSimApplication().getIntegrator(); // get integrator - check: better way to get integrator


            // time of next time event
            double tn;
            if (eventList.isEmpty())
                tn = ModelComponent.INFINITY;
            else
                tn = getTimeOfNextTimeEvent();

            if (i.continuousExist()) {    // make an integration step ------------------

                // find time to integrate (either timeStep or to next event).
                double tintegr = getSimApplication().getTime() + i.getStepWidth();
                if (tn <= tintegr && !eventList.isEmpty()) {
                    // integrate to next time event
                    try {
                        // make integration step
                        stateEventThrown = i.integrationStep(tn, tn - getSimApplication().getTime());
                        // if tn reached (no state event before) execute next time event
                        if (getSimApplication().getTime() >= tn) {
                            triggerNextTimeEvent();
                            eventThrown = true;
                        }
                    }
                    catch (ContSimulationException e) {
                        getSimApplication().errorOutput(e.getMessage());
                    }

                }
                else {
                    // integrate stepWidth long
                    try {
                        getSimApplication().debugOutput("simulation", " integration step for time " + tintegr, 2);
                        stateEventThrown = i.integrationStep(tintegr, i.getStepWidth());
                    }
                    catch (ContSimulationException e) {
                        getSimApplication().errorOutput(e.getMessage());
                    }
                }
            }    // end integration step ---------------------------------------

            // when no continuous part exists, only execute a time event
            else {

                getSimApplication().setTime(tn);
                getSimApplication().debugOutput("simulation", " event list: " + getEventListDebugString(), 1);
                if (!eventList.isEmpty()) {
                    triggerNextTimeEvent();
                    eventThrown = true;
                }
            } // end time event   ----------------------------------------------

            // fire all variable changes
            fireVariableChanges();

            // handle other threads
///*af
           if (!ted.isEmpty())
                ted.processAll();
//af*/
            getSimApplication().debugOutput("simulation", " stop step ----------------------------------------", 2);
        } // end stepLock

   } // end step


//event list and event triggerig -------------------------------------------------------------------

    /** sorted set storing the components with events according to event time (tn) and select order.
     * A synchronized TreeSet is used for this purpose.
     * Sorting is by class TimeEventSortingCriteria which sorts based on tn first and selectOrder second.
     */
    private SortedSet eventList =
                Collections.synchronizedSortedSet(new TreeSet(new TimeEventSortingCriteria()));

    /*
     * Comparator class for sorting of time events in event list.
     * Sorting is by class TimeEventSortingCriteria which sorts based on tn first and selectOrder second.
     */
    private class TimeEventSortingCriteria implements Comparator {

        /**
         * Compare o1 and o2 based on tn and selectOrder.
         * @return  -1 if o1 has smaller tn, 1 if o2 has smaller tn,
         *          -1 if o1 and o2 have equal tn but o1 has bigger selectOrder
         *          1 otherwise
         */
        public int compare(Object o1, Object o2) throws ClassCastException{
            if (o1 == o2) {
                return 0;
            } else {
                try {
                    TimeEventComponent tec1 = (TimeEventComponent)o1;
                    TimeEventComponent tec2 = (TimeEventComponent)o2;
                    if (tec1.getTN() < tec2.getTN()) {
                        return -1;
                    } else if (tec2.getTN() < tec1.getTN()) {
                        return 1;
                    } else if (tec1.getSelectOrder() > tec2.getSelectOrder()) {
                        return -1;
                    } else if (tec1.getSelectOrder() < tec2.getSelectOrder()) {
                        return 1;
                    } else if (tec1.getArtificialID()<tec2.getArtificialID()) {
                        return -1;
                    } else {
                        return 1;
                    }
                } catch (ClassCastException e) {
                    getSimApplication().errorOutput("Tried to put non TimeEventComponent into eventList "
                        + e.toString());
                    throw (e);
                }
            }
        }

        /** Not used. Always returns false */
        public boolean equals(Object obj) {
            return false;
        }

    } // end TimeEventSortingCriteria

    /** Adds an event to the event list */
    final private void addEvent(TimeEventComponent tEvt) {
        eventList.add(tEvt);
    }

    /** Registers an event to the event list */
    final public void scheduleEvent(TimeEventComponent tEvt, double time) {
        eventList.remove(tEvt);
        if (time < getSimApplication().getTime()) {
            getSimApplication().errorOutput(">> past event:" + tEvt.getTN() + "<" + getSimApplication().getTime() + "(" + tEvt.toString() + ")");
        } else if (time != ModelComponent.INF) {
            tEvt.setTN(time);
            addEvent(tEvt);
         }
    }

    /** Removes an event from the event list */
    final public void unscheduleEvent(TimeEventComponent tEvt) {
        eventList.remove(tEvt);
        tEvt.setTN(ModelComponent.INF);
    }

    /** Removes an event from the event list */
    final public void removeAllEvents() {
        eventList.clear();
    }

    /** Tests if the model m is the first in the event list. */
    final public boolean isNextToFire(TimeEventComponent m) {
        if (eventList.isEmpty()) return false;
        TimeEventComponent next= (TimeEventComponent)eventList.first();
        return next==m;
    }

    /** Gets the next model to fire which is the first in the event list. */
    final public TimeEventComponent getNextTimeEvent() {
        if (eventList.isEmpty()) return null;
        return (TimeEventComponent)eventList.first();
    }

    /** Triggers the next event in the event list by calling the triggerEvent of the active model */
    final private void triggerNextTimeEvent() {
        TimeEventComponent next = (TimeEventComponent)eventList.first();
        eventList.remove(next);
        getSimApplication().setTime(next.getTN());
        getSimApplication().debugOutput("simulation", " time event in " + next.getLabel() + " at " + next.getTN(), 1);
        next.triggerEvent();
        incNrEvents();
    }

    /** Gets the time of the next Time event in event list
     *  if no events in list, then return INFINITY */
    final public double getTimeOfNextTimeEvent() {
        if (eventList.isEmpty())
            return ModelComponent.INFINITY;
        else {
            TimeEventComponent next= (TimeEventComponent)eventList.first();
            return next.getTN();
        }
    }

    /** Gets time of next simulation step */
    final public double getNextSimulationTime () {
        if (!inputEventList.isEmpty()) {
            return getSimApplication().getTime();
        }
        double tn = getTimeOfNextTimeEvent();
        if (getSimApplication().getIntegrator().continuousExist()) {
            double tintegr = getSimApplication().getTime() + getSimApplication().getIntegrator().getStepWidth();
            if (tn <= tintegr)
                return tn;
            else
                return tintegr;
        }
        else
            return tn;
    }

    /** Generates a debug output for the event list  */
    public String getEventListDebugString() {
        final int MAX_NR_OUTPUT = 10;
        // putting out the event list elements
        String buf = "";
        Iterator e = eventList.iterator();
        int counter = 0;
        while (e.hasNext() && counter < MAX_NR_OUTPUT) {
            TimeEventComponent tec = (TimeEventComponent) e.next();
    		buf = buf + tec.getLabel() + ":" + tec.getTN() + "->";
        }
        buf = buf + "*]";
        return buf;
    }


// input event list and input event triggerig ------------------------------------------------------

    /** ArrayList of input events, not sorted, FIFO */
    private List inputEventList= Collections.synchronizedList(new ArrayList());

    /** Adds an event to the input event list */
    final public void addInputEvent(InputEventComponent iEvt) {
        inputEventList.remove(iEvt);
        inputEventList.add(iEvt);
    }

    /** Removes an event from the event list */
    final public void removeInputEvent(InputEventComponent iEvt) {
        inputEventList.remove(iEvt);
    }

    /** Removes all events from the inputevent list */
    final public void removeAllInputEvents() {
        inputEventList.clear();
    }

    /** Registers an event by adding the inputEventComponent model to the inputEvent list */
    final public void registerInputEvent(InputEventComponent s) {
        addInputEvent(s);
    }

    /** Unregisters an event by removing the model from the event list */
    final public void unregisterInputEvent(InputEventComponent s) {
        removeInputEvent(s);
    }

    /** Gets and removes the first in the event list. Returns null if empty*/
   final  public InputEventComponent nextInputEventComponent() {
        if (! inputEventList.isEmpty()){
            InputEventComponent s = (InputEventComponent)inputEventList.get(0);
            inputEventList.remove(s);
            return s;
        } else {
            return null;
        }
    }

    /** Triggers the next event in the event list by calling the triggerEvent of the inputEventComponent */
    final private void triggerNextInputEvent() {
        InputEventComponent n = nextInputEventComponent();
        getSimApplication().debugOutput("simulation", " input event " + n.toString(), 2);
        n.triggerEvent();
    }


//variable change events    ------------------------------------------------------------------------

    /** sorted set storing the registries for the variable change events.
     * Variable change registries hold the variable and the listener. Those are sorted in
     * priority order in this sorted set.
     * Sorting is by class VariableChangeRegistrySortingCriteria.
     */
    private SortedSet variableChanges =
        Collections.synchronizedSortedSet(new TreeSet(new VariableChangeRegistrySortingCriteria()));

    /*
     * Comparator class for sorting of variable change registries in variable change event list.
     * Sorting is by class VariableChangeRegistrySortingCriteria which sorts based on priortiy
     * of the listeners. Does not allow duplicates.
     */
    private class VariableChangeRegistrySortingCriteria implements Comparator {

        /**
         * Compare o1 and o2 based on tn and selectOrder.
         * @return  1 if o1 has smaller tn, -1 if o2 has smaller tn,
         *          1 if o1 and o2 have equal tn but o1 has bigger selectOrder
         *          -1 otherwise
         */
        public int compare(Object o1, Object o2) throws ClassCastException{
            try {
                VariableChangeRegistry vcr1 = (VariableChangeRegistry)o1;
                VariableChangeRegistry vcr2 = (VariableChangeRegistry)o2;
                if (vcr1.getVariable() == vcr2.getVariable() && vcr1.getListener() == vcr2.getListener()) {
                    return 0; // they are equal events and not stored in the set.
                } else if (vcr1.getListener().getPriority() <= vcr2.getListener().getPriority()) {
                    return 1;
                } else  {
                    return -1;
                }
            } catch (ClassCastException e) {
                getSimApplication().errorOutput("Tried to put non VariableChangeRegistry into variableChanges set "
                    + e.toString());
                throw (e);
            }
        }

        /** Not used. Always returns false */
        public boolean equals(Object obj) {
            return false;
        }

    } // end TimeEventSortingCriteria


    /** Registers a variable change event by adding a ChangeRegistry into the changeEventList */
    public void registerChangeEvent(VariableChangeRegistry cr) {
        getSimApplication().debugOutput("simulation", "Register: " + cr.toString(), 3);
        variableChanges.add(cr);
    }

    /** Fires a variable change event by firing a  change event
     *   for the first ChangeRegistry in the the changeEventList */
    public void fireVariableChanges() {
        SortedSet currentChanges;
        ChangeRegistry cr;
        while (!variableChanges.isEmpty()) {
            currentChanges = variableChanges;
            variableChanges = Collections.synchronizedSortedSet(
                                        new TreeSet(new VariableChangeRegistrySortingCriteria()));
            Iterator it = currentChanges.iterator();
            getSimApplication().debugOutput("simulation", "-------------------------", 3);
            while (it.hasNext()) {
                cr = (ChangeRegistry )it.next();
                getSimApplication().debugOutput("simulation", "Execute " + cr.toString(), 3);
                cr.fireChangeEvent();
            }
        }
    }

    /** Removes all registered variable change events */
    public void removeAllChangeEvents() {
        variableChanges.clear();
    }

    public void fireDiscreteAndContinuousVariableChanges() {
        getSimApplication().getIntegrator().signalVariableChanges();
        fireVariableChanges();
    }

// -------------------------------------------------------------------------------------------------
// interface ModelComponentEventListener -----------------------------------------------------------
// -------------------------------------------------------------------------------------------------

    /** Handle the addition of a model element to a model component */
    public void modelComponentAdded(ModelComponentEvent e) {
        handleAdditionOfModelComponent(e.getModelComponent(), e.getSubModelComponent());
    }

    /** Handle the removal of a model element to a model component */
    public void modelComponentRemoved(ModelComponentEvent e) {
        handleRemovalOfModelComponent(e.getModelComponent(), e.getSubModelComponent());
    }

    public void modelComponentEnabled (ModelComponentEvent e) {
        handleEnablingOfModelComponent(e.getModelComponent());
    }

    public void modelComponentGotContinuous(ModelComponentEvent e){
        handleGettingContinuousOfModelComponent(e.getModelComponent());
    }

    public void modelComponentGotDiscrete(ModelComponentEvent e){
        handleGettingDiscreteOfModelComponent(e.getModelComponent());
    }

    public void modelComponentDisabled (ModelComponentEvent e) {
        handleDisablingOfModelComponent(e.getModelComponent());
    }

    public void modelComponentChanged (ModelComponentEvent e) {
        handleChangeOfModelComponent(e.getModelComponent());
    }

    protected void handleAdditionOfModelComponent (ModelComponent model, ModelComponent element) {
        // handle all subcomponents
        Iterator subElements = element.getModelComponents().iterator();
        while (subElements.hasNext()) {
            ModelComponent subElem = ((ModelComponent) (subElements.next()));
            handleAdditionOfModelComponent(element, subElem);
        }

        // add simulator as model component listener to element
        element.addModelComponentEventListener(this);

        // register the continuous variables with the integrator
        if (element instanceof DoubleVariable && ((DoubleVariable)element).isContinuous()) {
            getSimApplication().getIntegrator().add((DoubleVariable)element);
        }

        // register the continuous equaitons in continuous mode with the integrator
        if (element instanceof ContEquation) {
            if (((ContEquation)element).isContinuousMode()) {
                getSimApplication().getIntegrator().add((ContEquation)element);
            }
        }

        // register the state event component with the integrator
        if (element instanceof StateEventComponent) {
            getSimApplication().getIntegrator().add((StateEventComponent)element);
        }

        // register the connections in continuous mode with the integrator
        if (element instanceof Connection && ((Connection)element).isContinuousMode()) {
            getSimApplication().getIntegrator().add((Connection)element);
        }

        // register enabled components
        if (element.isEnabled()) {
            registerEnabledComponent(element);
        } else {
            deregisterDisabledComponent(element);
        }
    }

    protected void handleRemovalOfModelComponent (ModelComponent model, ModelComponent element) {

        Iterator subElements = element.getModelComponents().iterator();
        while (subElements.hasNext()) {
            ModelComponent subElem = ((ModelComponent) (subElements.next()));
            handleRemovalOfModelComponent(element, subElem);
        }
        element.removeModelComponentEventListener(this);


        if (element instanceof DoubleVariable) {
        getSimApplication().getIntegrator().remove((DoubleVariable)element);
        }

        if (element instanceof ContEquation) {
            getSimApplication().getIntegrator().remove((ContEquation)element);
        }

        if (element instanceof Connection) {
            getSimApplication().getIntegrator().remove((Connection)element);
        }

        //deregisterDisbledComponent
        deregisterDisabledComponent(element);
    }

    protected void handleEnablingOfModelComponent (ModelComponent element) {
        // first handle enabling of all subcomponents
        Iterator subElements = element.getModelComponents().iterator();
        while (subElements.hasNext()) {
            ModelComponent subElem = ((ModelComponent) (subElements.next()));
            handleEnablingOfModelComponent(subElem);
        }
        registerEnabledComponent(element);
    }

    protected void registerEnabledComponent(ModelComponent element) {

        if (element instanceof Variable) {
            getSimApplication().getIntegrator().enable((Variable)element);
        }

        if (element instanceof ContEquation) {
            getSimApplication().getIntegrator().enable((ContEquation)element);
        }

        if (element instanceof Connection) {
            getSimApplication().getIntegrator().enable((Connection)element);
        }

        // signal variable changes for added discrete variables.
        if (element instanceof Variable) {
            if  (
                  (! (element instanceof DoubleVariable) ) ||
                  ( element instanceof DoubleVariable && !((DoubleVariable)element).isContinuous() )
                ) {
                getSimApplication().debugOutput("simulation",
                                "signaling variable change because of enabling variable" + element.getLabel(),
                                4);
                ((Variable)element).signalVariableChange();
            }
        }

        // schedule added time event components in the event list.
        if (element instanceof TimeEventComponent) {
            TimeEventComponent timer = (TimeEventComponent) element;
            getSimApplication().debugOutput("simulation",
                "adding time event component " + element.getLabel() + " from event list because enabled ",
                4);
            addEvent(timer);
        }

        // register the animated view objects with the animator
        if (element instanceof AnimatedView) {
            getSimApplication().getAnimator().addView((AnimatedView)element);
        }
    }

    protected void handleDisablingOfModelComponent (ModelComponent element) {
        // first handle disabling of all subcomponents
        Iterator subElements = element.getModelComponents().iterator();
        while (subElements.hasNext()) {
            ModelComponent subElem = ((ModelComponent) (subElements.next()));
            handleEnablingOfModelComponent(subElem);
        }

        deregisterDisabledComponent(element);
    }

    protected void deregisterDisabledComponent(ModelComponent element) {
        if (element instanceof Variable) {
            getSimApplication().getIntegrator().disable((Variable)element);
        }

        if (element instanceof ContEquation) {
            getSimApplication().getIntegrator().disable((ContEquation)element);
        }

        if (element instanceof Connection) {
            getSimApplication().getIntegrator().disable((Connection)element);
        }

        // remove variable changes for disabled discrete variables.
        if (element instanceof Variable) {
            if  (
                  (! (element instanceof DoubleVariable) ) ||
                  ( element instanceof DoubleVariable && !((DoubleVariable)element).isContinuous() )
                ) {
                getSimApplication().debugOutput("simulation",
                                "removing variable change because of disabling variable" + element.getLabel(),
                                3);
                //((Variable)element).removeVariableChange();
            }
        }

        // remove time event components in the event list.
        if (element instanceof TimeEventComponent) {
            TimeEventComponent timer = (TimeEventComponent) element;
            getSimApplication().debugOutput("simulation",
                "removing time event component " + element.getLabel() + " from event list because disabled ",
                3);
            unscheduleEvent(timer);
        }

        // register the animated view objects with the animator
        if (element instanceof AnimatedView) {
            getSimApplication().getAnimator().removeView((AnimatedView)element);
        }
    }

    protected void handleGettingContinuousOfModelComponent(ModelComponent element) {
        if (element instanceof DoubleVariable) {
            getSimApplication().getIntegrator().add((DoubleVariable)element);
        }

        if (element instanceof ContEquation) {
            getSimApplication().getIntegrator().add((ContEquation)element);
        }

        if (element instanceof StateEventComponent) {
            getSimApplication().getIntegrator().add((StateEventComponent)element);
        }

        if (element instanceof Connection) {
            getSimApplication().getIntegrator().add((Connection)element);
        }
    }

    protected void handleGettingDiscreteOfModelComponent(ModelComponent element) {
        if (element instanceof DoubleVariable) {
            getSimApplication().getIntegrator().remove((DoubleVariable)element);
        }

        if (element instanceof ContEquation) {
            getSimApplication().getIntegrator().remove((ContEquation)element);
        }

        if (element instanceof StateEventComponent) {
            getSimApplication().getIntegrator().remove((StateEventComponent)element);
        }

        if (element instanceof Connection) {
            getSimApplication().getIntegrator().remove((Connection)element);
        }
    }

    protected void handleChangeOfModelComponent (ModelComponent element) {
        if (element instanceof Variable) {
            getSimApplication().getIntegrator().change((Variable)element);
        }

        if (element instanceof ContEquation) {
            getSimApplication().getIntegrator().change((ContEquation)element);
        }

        if (element instanceof Connection) {
            getSimApplication().getIntegrator().change((Connection)element);
        }
    }

// -------------------------------------------------------------------------------------------------
// Diverses    -------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------

//serialization (for later use) --------------------------------------------------------------------

    private void writeObject(java.io.ObjectOutputStream out)
    throws IOException  {
  	    // for later use
    }

    private void readObject(java.io.ObjectInputStream in)
    throws IOException, ClassNotFoundException  {
  	    // for later use
    }

}   // end Simulator --------------------------------------------------------------------------------




