
//Project:      mosaic kernel
//Version:      0.3
//Copyright:    Copyright (c) 1998
//Author:       Herbert Praehofer, Johannes Sametinger, Alois Stritzinger
//Affiliation:  Johannes Kepler University, Linz, Austria

// -----------------------------------------------------------------------------------------------------------
// Class Timer      ------------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------
// programmed by: hp
// last modified by: hp, 14.12.98
// documented by: hp, 14.12.98


package mosaic.sim.kernel.simevent;

import mosaic.sim.kernel.control.*;
import java.beans.*;
import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import mosaic.sim.kernel.AbstractModelComponent;
import mosaic.sim.util.artificialidentifier.ArtificialIDGiver;


/**
* Class implementing a Timer to trigger time events.
* <p>
* Implements the interface TimeEventComponent which specifies the protocol used by the Simulator
* to schedule time events.
* Provides scheduling functions like activateAt, activateIn, delay, stop, etc. used in the Models.
* Has a timeEvent which is signaled to the TimeEventListeners when the event is executed.
* <p>
* Use this bean as a component to schedule events. Set the time of signaling by
* using activateIn, activate, or activateAt. Add listeners to the time event using
* addTimeEventListener. Listeners are notified by a call to method processSimEvent.
* (Works similar as a clock which rings at the time set.)
*
* @see sim.kernel.TimeEvent
* @see sim.kernel.TimeEventListener
* @see sim.kernel.ModelComponent
*/

public class Timer  extends AbstractModelComponent
                    implements TimeEventComponent {


// Constructors ----------------------------------------------------------------------------------------------

  /** Default constructor Timer */
  public Timer() {
      this("T");
  }

  /** Timer constructor with label*/
  public Timer(String iLabel) {
    super(iLabel);
    setForeground(Color.red);
    setFont(new Font("SansSerif", Font.PLAIN, 10));
    setInitTime (INF);
  }

// -----------------------------------------------------------------------------------------------------------
// interface ArtificalIDOwner --------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------
  private final int artificialID=ArtificialIDGiver.getID();
  public final int getArtificialID() {
    return artificialID;
  }


// -----------------------------------------------------------------------------------------------------------
// interface Comparable --------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------

  /** compares this timer with another timer c based on tn and selectOrder */
  public boolean before(Comparable c) {
    if (c instanceof TimeEventComponent) {
      TimeEventComponent s= (TimeEventComponent)c;
      if (getTN()==s.getTN())
        return getSelectOrder()>s.getSelectOrder();
      else return getTN()<=s.getTN();
    }
    return false; //use exception!!
  }



// -----------------------------------------------------------------------------------------------------------
// interface  ModelComponent     -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------

// reset -----------------------------------------------------------------------------------------------------

  /** Resets the Timer by setting properties stopped, active and event registration at initTime */
  public void reset() {
    super.reset();
    setStopped (false);
    setActive (false);
    activateAt(getInitTime());
  }


// paint -----------------------------------------------------------------------------------------------------

  /* Paints the timer */
  public void paintModelComponent(Graphics g) {
      int h = getSize().height;
      int w = getSize().width;
      g.setColor(getForeground());
      g.drawOval(0, 0, w-1, h-1);
      g.drawLine(w/2, h/2, w-1, 0);
      g.setFont(getFont());
      FontMetrics fm = g.getFontMetrics();
      g.setColor(getForeground());
      String lbl = getLabel();
      if (lbl == null) {
          lbl = "V";
      }
    g.drawString(lbl,2,h-fm.getMaxDescent());
  }

//size: ------------------------------------------------------------------------------------------------------

  /** Gets the minimum size for the LayoutManager */
  public Dimension getMinimumSize() {
    return new Dimension(5,5);
  }

  /** Gets the preferred size for the LayoutManager */
  public Dimension getPreferredSize() {
    return new Dimension(16,12);
  }


//info: -----------------------------------------------------------------------------------------------------

  /** Generates a debug string for the timer */
  protected String getDebugText() {
      String tnString = "";
      if (getTN() == INF)
         tnString = "INF";
      else
         tnString = tnString + getTN();
    return "tn=" + tnString;
  }


// -----------------------------------------------------------------------------------------------------------
// interface TimeEventComponent ------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------

// next event time -------------------------------------------------------------------------------------------

  /** variable to hold the time tn of the next event for the timer */
  protected double tn= 0;

  /** Gets the time tn of the next event for the timer */
  public double getTN() {
      return tn;
  }

  /** Sets the time tn of the next event for the timer */
  public void setTN(double newTime) {
    double oldTN = tn;
    tn = newTime;
  }


// last event time -------------------------------------------------------------------------------------------

  /** variable to hold the time tl of the last event for the timer */
  protected double tl= 0;

  /** Gets the time tl of the last event for the timer */
  public double getTL() { return tl; }

  /** Sets the time tl of the last event for the timer */
  public void setTL(double newTime) {
    double oldTL = tn;
    tl = newTime;
  }


// select order  ---------------------------------------------------------------------------------------------

  /** variable to hold the selectOrder for the timer which is used as a priority
  * for tie breaking between several timer with equal next event time.
  */
  protected int selectOrder = 0;

  /** Gets the select order of the timer */
  public int getSelectOrder() {
      return selectOrder;
  }

  /** Sets the select order of the timer */
  public void setSelectOrder(int newOrder) {
    int oldSelectOrder = selectOrder;
    selectOrder= newOrder;
  }

// triggerEvent  ---------------------------------------------------------------------------------------------

  /** Triggers an event, that is, it executes the internal transition function */
    public void triggerEvent() {
    setTL(getTime());
    unregisterEvent();
    timeEvent.setTime(getTime());
    Vector timeEventListenersCopy = (Vector)(timeEventListeners.clone());
    for (int idx = 0; idx < timeEventListenersCopy.size(); idx++) {
          ((SimEventListener)timeEventListenersCopy.get(idx)).processSimEvent (timeEvent);
    }
  }


// timeEvent -------------------------------------------------------------------------------------------------

  /** Vector of TimeEventListeners */
  Vector timeEventListeners = new Vector();

  /** single TimeEvent in Timer */
  mosaic.sim.kernel.simevent.TimeEvent timeEvent = new TimeEvent(this);

  /** Add SimEventListener */
  public void addSimEventListener (mosaic.sim.kernel.simevent.SimEventListener l) {
      if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
        getSimApplication().getIntegrator().stateEventOccurred = true;
        return;
      }
      timeEventListeners.add(l);
      l.addSimEventComponent(this);
  }

  /** Remove SimEventListener */
  public void removeSimEventListener (mosaic.sim.kernel.simevent.SimEventListener l){
     if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
       getSimApplication().getIntegrator().stateEventOccurred = true;
       return;
     }
     timeEventListeners.remove(l);
      l.removeSimEventComponent(this);
  }

  /** Add TimeEventListener */
  public void addTimeEventListener (mosaic.sim.kernel.simevent.TimeEventListener l) {
      if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
        getSimApplication().getIntegrator().stateEventOccurred = true;
        return;
      }
      addSimEventListener(l);
  }

  /** Remove TimeEventListener */
  public void removeTimeEventListener (mosaic.sim.kernel.simevent.TimeEventListener l){
      if (getSimApplication() != null && getSimApplication().getIntegrator().testForStateEvents) {
        getSimApplication().getIntegrator().stateEventOccurred = true;
        return;
      }
      removeSimEventListener(l);
  }



// -----------------------------------------------------------------------------------------------------------
// scheduling functions used in the model specifications ----------------------------------------------
// -----------------------------------------------------------------------------------------------------------

    /** Activates the Timer immediately (after 0 time units). */
    public void activate () {
        registerEvent(getTime());
    }

    /** Activates the Timer at time <code>time</code> */
    public void activateAt (double time) {
        registerEvent(time);
    }

    /** Activates the Timer in  <code>deltaT</code> time units */
    public void activateIn (double deltaT) {
        try {
            registerEvent(getTime()+deltaT);  // might cause overflow
        } catch (Exception e) {}
    }

    /** Passivates the Timer by removing it from the event list */
    public void passivate() {
        unregisterEvent ();
    }


// stop and restart  -----------------------------------------------------------------------------------------
// (a stopped Timer can be restarted again with the remaining time)

  /** variable to hold the remaining time when stopped */
  protected double deltaT = 0;

  /** boolean to signify if the timer has been stopped */
  protected boolean stopped = false;

  /** Stops the timer and saves the remaining time */
  public void stop () {
      deltaT = getTN() - getTime();
      setStopped (true);
      passivate ();
  }

  /** Restarts the timer with remaining time */
  public void restart () {
      if (isStopped()) {
         setStopped (false);
         activateIn (deltaT);
      }
  }

  /** Tests if the timer has been stopped */
  public boolean isStopped () {
      return stopped;
  }

  /** Sets the stopped flag of */
  public void setStopped (boolean newStopped) {
    stopped= newStopped;
  }


//property active and passive --------------------------------------------------------------------------------

  /** variable to hold the property active */
  protected boolean active= false;

  /** Tests if the timer is active */
  public boolean isActive() {
      return active;
  }

  /** Sets the active property of the timer */
  protected void setActive(boolean a) {
      boolean oldActive= active;
      active= a;
  }

  /** Tests if the timer is passive (! active) */
  public boolean isPassive() {
      return !active;
  }


/// property initTime for first simulation step  -------------------------------------------------------------

  /** variable to hold the initial event time for the timer. Is INF by default.  */
  protected double initTime = INF;

  /** Gets the initial event time for the timer. */
  public double getInitTime() {
      return initTime;
  }

  /** Sets the initial event time for the timer. */
  public void setInitTime(double newInitTime) {
      initTime= newInitTime;
  }


// -----------------------------------------------------------------------------------------------------------
// implementation --------------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------

// event registration ----------------------------------------------------------------------------------------

  /** Registers an event for time <code>time</time>, that is, it is inserted into the event list. */
  protected void registerEvent(double time) {
    if (isStopped()) setStopped (false);
//    setTN(time);
    if (time >= getTime()) {
        setActive (true);
        printDebugOutput("simulation", "try to schedule event : " + this.getLabel() + " at time " + time, 2);
        getSimApplication().getSimulator().scheduleEvent(this, time);
    }
    else {
      printErrorOutput(">> tried to schedule past event:" + time + "<"
                          + getTime() + " in timer " + toString() + "!");
      unregisterEvent();
    }
  }

  /** Unregisters a Timer, that is, it is removed from the event list. */
  protected void unregisterEvent() {
    getSimApplication().getSimulator().unscheduleEvent(this);
    setActive(false);
  }

  /** Tests if the timer is the next in the event list. */
  public boolean isNextToFire() {
    return getSimApplication().getSimulator().isNextToFire(this);
  }



//to string --------------------------------------------------------------------------------------------------

  /** Makes a string representation of the timer */
  public String toString() {
      String tnString = "";

      if (getTN() == INF)
         tnString = "INF";
      else
         tnString = tnString + getTN();
    return getLabel() +
           "[tn=" + tnString +
           "]";
  }



}  // end of Timer  ------------------------------------------------------------------------------------------

