
//Project:      mosaic kernel
//Version:      0.3
//Copyright:    Copyright (c) 1998
//Author:       Herbert Praehofer, Johannes Sametinger, Alois Stritzinger
//Affiliation:  Johannes Kepler University, Linz, Austria

// -----------------------------------------------------------------------------------------------------------
// class BasicVariableObserver             -------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------
// programmed by: hp
// last modified by: hp, 14.12.98
// documented by: hp, 8.4.99


package mosaic.sim.kernel.variable;

import java.util.*;
import mosaic.sim.kernel.*;
import java.beans.*; 

/**
* Base class for  observers of variable changes.
* Observers are special VariableChangeListeners with no feedback to the changed component
* but only listen to them, e.g. Views.
* <p>
* Has value object.
* Special observers should be built by specializing method <code> setValue </code>
* <p>
* Can have ValueFilter object to filter values received.
* @see   sim.kernel.output.*
* @see   sim.kernel.statistics.*
*/
abstract public class BasicVariableObserver
                extends AbstractModelComponent
				implements VariableChangeListener
{
  /** Default constructor ValueView */
  public BasicVariableObserver() {
      setBackground(java.awt.Color.lightGray);
      value = null;
      setLabel(" "); 
  }

  /** Constructor initializing label */
  public BasicVariableObserver(String label) {
      super (label);
  }

// value -----------------------------------------------------------------------------------------------------

  /** variable to hold the Object value from the variable change event*/
  protected Object value;

  /** Gets the Object value from the variable change event*/
  public Object getValue () {
      return value;
  }

  /** Sets the Object value from the variable change event*/
  public void setValue (Object newVal) {
      value = newVal;
  }


// ValueFilter -----------------------------------------------------------------------------------------------
// ValueFilters are used to filter and transform the received values

  /** variable to hold the ValueFilter strategy object  */
  ValueFilter valueFilter;

  /** Gets the valueFilter */
  public ValueFilter getValueFilter () {
      return valueFilter;
  }

  /** Sets the valueFilter */
  public void setValueFilter (ValueFilter newValueFilter) {
      valueFilter = newValueFilter;
  }

  /** Filters the received values by calling filterValue of its ValueFilter if exists. */
  final public  Object filterValue (Object value) {
      if (valueFilter == null)
          return value;
      else
          return valueFilter.filterValue (value); 
  }

  /** Tests the received values by calling testValue of its ValueFilter if exists. */
  final public boolean testValue (Object value) {
      if (valueFilter == null)
          return true;
      else
          return valueFilter.testValue (value);
  }

// variableChange --------------------------------------------------------------------------------------------

  /** Reacts to a variableChange event by first filtering and then setting the value (method setValue).*/
  public void variableChange(VariableChangeEvent e) {
      if (e==null) return;
      if (testValue (e.getValue()))
        setValue(filterValue(e.getValue()));
  }

  /** Reacts to a propertyChange event by calling variableChange. */
  public void propertyChange(PropertyChangeEvent e) {
      if (e instanceof VariableChangeEvent)
          variableChange((VariableChangeEvent) e);
  }


// listenerVariables --------------------------------------------------------------------------------------------

  /** Add the variable to the list of variables this observer listens to. */
  public void addListenerVariable(Variable v) {
      if (! listenerVariables.contains(v)) {
         listenerVariables.add(v);
      }
   }

   /**Remove the variable from the list of variables this observer listens to. */
   public void removeListenerVariable (Variable v) {
      listenerVariables.remove(v);
   }

   /** vector of variables this observer listens to. */
   protected Vector listenerVariables = new Vector();

   /** Gets the vector of variables this observer listens to.*/
   public Vector getListenerVariables () {
      return listenerVariables;
   }

   /** Gets nth variable this observer listens to.*/
   public Variable getNthListenerVariable (int i) {
      if (i >= listenerVariables.size())
         return null;
      else
         return ((Variable) listenerVariables.get(i));
   }

// priority ----------------------------------------------------------------------------------------

    /** variable to hold the priority for handling this listener in handling variable change events. */
    protected int priority = Integer.MAX_VALUE;

    /**
     * Gets the priority for the handling of variable change events by the simulator (the higher
     * the priority, the earlier it will be handled.)
     */
    public final int getPriority() {
        return priority;
    }

}

