/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: NeuronOutputSupport.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.8 $
 *  Last changed at: $Date: 2002/05/16 20:43:26 $
 *  Last changed by: $Author: rene $
 *****************************************************************************/

package mosaic.sim.neuron;

import java.util.*;
import mosaic.sim.kernel.*;

/** This is a helper class for objects that want to emit spike events. It
 *  implements everything required by the NeuronOutput interface and also
 *  offers the method fireSpikeEvent for sending a spike event to all
 *  registered NeuronInput listeners.
 */
public class NeuronOutputSupport extends BasicActiveModel
		implements NeuronOutput {
	/** This is the list of NeuronInput objects which should be notified
	 *  whenever this object emits a spike event.
	 *  It should only contain objects implementing the NeuronInput interface,
	 *  which will typically be Synapse objects.
	 */
	protected ArrayList outputs;

	/** Sends a spike event to every registered NeuronInput listener object.
	 */
	protected void fireSpikeEvent() {
		Debug.output(Debug.LEVEL_INFO, "NeuronOutputSupport.fireSpikeEvent for " + label
			+ " at time " + getTime() + "\n");
		for (int i=0; i<outputs.size(); i++)
			((NeuronInput) outputs.get(i)).spikeEvent(this);
		Debug.output(Debug.LEVEL_INFO, "fireSpikeEvent complete\n");
	}

	/** Default constructor, initializes the output list.
	 */
	public NeuronOutputSupport() {
		outputs = new ArrayList();
	}

	/** Returns the list of currently registered NeuronInput objects that should
	 *  receive the spike events sent by this object.
	 */
	public NeuronInput[] getOutputs() {
		return (NeuronInput[]) outputs.toArray();
	}

	/** Sets the list of currently registered NeuronInput objects that should
	 *  receive the spike events sent by this object.
	 *  This removes all objects that were registered before, so this method
	 *  should only be used during initialization.
	 *  For adding and removing NeuronInput objects without removing the other
	 *  registered objects, use the methods addOutput / removeOutput.
	 *  @see #addOutput
	 *  @see #removeOutput
	 */
	public void setOutputs(NeuronInput[] receivers) {
		Debug.assertCond(receivers != null, "NeuronOutputSupport.setOutputs",
			"receivers can not be null");
		outputs = new ArrayList();
		for (int i=0; i<receivers.length; i++)
			outputs.add(receivers[i]);
	}

	/** Adds a single NeuronInput object to the list of objects that should
	 *  receive spike events sent by this object.
	 */
	public void addOutput(NeuronInput receiver) {
		outputs.add(receiver);
	}

	/** Removes a single NeuronInput object from the list of objects that should
	 *  receive spike events sent by this object.
	 *  @return true when the given object was found and could be removed and
	 *           false when the given object could not be found in the current
	 *           list.
	 */
	public boolean removeOutput(NeuronInput receiver) {
		return outputs.remove(receiver);
	}
}
