/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: PLFunction.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.12 $
 *  Last changed at: $Date: 2002/05/16 20:43:26 $
 *  Last changed by: $Author: rene $
 *****************************************************************************/

package mosaic.sim.neuron;

 /** This class represents a piecewise linear mathematical function in a
  *  finite timeframe with a finite number of linear segments. The linear
  *  segments are described by their gradients and their begin times and the
  *  whole function is defined to start at value 0 and time 0 with gradient 0.
  *  All gradients[i] are valid between the times gradientTimes[i] and
  *  gradientTimes[i+1]. Therefore the whole function is 0 in the interval
  *  [0, gradientTimes[0]]. The last segment, starting at time
  *  gradientTimes[numGradients] with gradient[numGradients] is defined to end
  *  at that time when the function reaches the value 0. This time can be
  *  retrieved by using the method getEndTime();
  */
public class PLFunction {
	/** Aargh ! This is a hack !
	 */
	public class FloatIntPair {
		public float f;
		public int i;
	}

	/** The gradients of the linear segments.
	 */
	public double[] gradients;

	/** The begin times of the linear segments.
	 */
	public double[] gradientTimes;

	/** The default constructor initializes both arrays with the same length.
	 */
	public PLFunction(int numSegments) {
		gradients = new double[numSegments];
		gradientTimes = new double[numSegments];
	}

	/** Calculates and returns the time when the function reaches 0.
	 */
	public double getEndTime() {
		Debug.assertCond(gradients.length == gradientTimes.length,
			"PLFunction.getEndTime",
			"gradient and time arrays must have equal length");
		return getEndTime(0.0, gradients.length);
	}

	/** Calculates and returns the time when the function, starting with the
	 *  given startValue as offset and when restricted to the first numSegments
	 *  segments, reaches 0.
	 */
	public double getEndTime(double startValue, int numSegments) {
		if (numSegments == 0)
			return 0;

		/* Count the number of segments until only segments with a gradient of
		   0 are left. */
		// Go back all the last segments that had a gradient of 0.
		while (numSegments > 0 && gradients[numSegments-1] == 0)
			numSegments--;
		if (numSegments == 0)
		   return 0;
		/* Calculate the value of this function until the beginning of the last
		   linear segment. */
		double value = startValue, endTime;
		for (int i=0; i<numSegments-1; i++)
			value += gradients[i] *	(gradientTimes[i+1] - gradientTimes[i]);

		// and now calculate the time until the last segment reaches the value 0
		endTime = - value / gradients[numSegments-1];
		// if this value is negative, then the function does not reach 0 at the
		// end
		Debug.assertCond(endTime >= 0, "PLFunction.getEndTime",
			"endTime must be >= 0 (function does not reach 0)");
		return endTime + gradientTimes[numSegments-1];
	}

	/**
	 * @see #getFinishedSegments(double startValue, int numSegments, double time, boolean untilNow)
	 */
	public FloatIntPair getFinishedSegments(double time, boolean untilNow) {
		return getFinishedSegments(0.0, gradients.length, time, untilNow);
	}

	/** Returns the value of the function at the time of last gradient
	 *  change before the given simulation time (computing all linear
	 *  segments that have ended before the current simulation time). Therefore
	 *  the difference between the value returned by this method and the
	 *  current function value is only the last, currently valid linear
	 *  segment of the function.
	 *  The method will return the
	 *  index of the last linear segment that
	 *  has been used in this computation.
	 *
	 * @param startValue The start value to use.
	 * @param numSegments The number of segments that should be used.
	 * @param time The simulation time to be used.
	 * @param untilNow Specifies if segments that end exactly at the current
	 *                 simulation time should also be taken into account.
	 */
	 public FloatIntPair getFinishedSegments(double startValue, int numSegments,
			double time, boolean untilNow) {
		/*Debug.assert(time >= gradients[0], "PLFunction.getFinishedSegments",
			"Can not calculate past developement, given time is before " +
			"starting time of the function.");*/

		FloatIntPair ret = new FloatIntPair();
		ret.f = (float) startValue;
		for (ret.i=0;
			 (untilNow && ret.i < numSegments - 1 &&
			 gradientTimes[ret.i+1] <= time) ||
			 (!untilNow && ret.i < numSegments - 1 &&
			 gradientTimes[ret.i+1] < time); ret.i++) {
			ret.f += gradients[ret.i] *
				(gradientTimes[ret.i+1] - gradientTimes[ret.i]);
		}
		return ret;
	}

	/** Returns the value of the function at the given time, starting with the
	 *  given startValue and taking numSegments into account.
	 *
	 * @see #getFinishedSegments
(double, int, double, boolean)
	 * @param startValue The start value to use.
	 * @param numSegments The number of segments that should be used.
	 * @param time The simulation time to be used.
	 */
	public double getValueAtTime(double startValue, int numSegments,
		   double time) {
		FloatIntPair ret = getFinishedSegments(startValue, numSegments, time, false);
		// add the last segment (until the current time)
		/* If the above loop terminated because there are no more future changes
		   available, we assume that the last gradient is used until the current
		   time. If it terminated because the next segment would have exceeded
		   the current time, we too can assume that the last gradient is
		   currently valid. */
		if (ret.i < numSegments)
		   ret.f += gradients[ret.i] * (time - gradientTimes[ret.i]);
		return ret.f;
	}

	/** Returns the function as string. Only for debugging purposes.
	 */
	 public String toString() {
		Debug.assertCond(gradients.length == gradientTimes.length,
			"PLFunction.toString",
			"gradient and time arrays must have equal length");
		return toString(gradients.length);
	 }

	/** Returns the first numSegments linear segments of this function as
	 *  string. Only for debugging purposes.
	 */
	 public String toString(int numSegments) {
		String ret = new String();
		for (int i=0; i<numSegments; i++) {
			ret += gradients[i] + " at " + gradientTimes[i] + "\n";
		}
		return ret;
	 }
}
