/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: PLPotential.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.24 $
 *  Last changed at: $Date: 2002/05/16 20:43:26 $
 *  Last changed by: $Author: rene $
 *****************************************************************************/

package mosaic.sim.neuron;

import mosaic.sim.kernel.*;
import mosaic.sim.kernel.simevent.*;

/** This class represents a neuron potential described by a piecewise linear
 *  mathematical function.
 */
public class PLPotential extends Potential {
	/** This function specifies the future potential development, computed from
	 *  the received synapse response functions (by mergeSynapseResponse).
	 */
	protected PLFunction changes;

	/** This is only a buffer for the mergeSynapseResponse method, used for the
	 *  enhanced merge sort in there. It is here to avoid too many dynamic
	 *  memory allocations.
	 */
	protected PLFunction buffer;

	/** The value of the Neuron potential at the (simulation) time of the last
	 *  update / merge. It is only updated by the mergeSynapseRespnse method.
	 */
	//protected float lastValue;

	/** The current number of linear segments in the piecewise linear potential
	 *  function 'changes'. It is needed because changes (and buffer) are
	 *  allocated with a maximum size (so that they do not need to be
	 *  re-allocated during simulation) and the real (valid) number of linear
	 *  segments currently describing the potential function will usually be
	 *  smaller then this.
	 *  @see #changes
	 */
	protected int numSegments;

	/** Initializes lastValue and numSegments with 0 and allocates changes and
	 *  buffer with maxSegements linear segments.
	 */
	public PLPotential(int maxSegments) {
		changes = new PLFunction(maxSegments);
		buffer = new PLFunction(maxSegments);
		//lastValue = 0;
		numSegments = 0;
	}

	/** Returns the current potential value (at the current simulation time). */
	public double getCurrentPotential() {
		return changes.getValueAtTime(0.0, numSegments, getTime());
	}

	/** Merges the synaptic response of an input synapse (probably caused by a
	 *  spike received by that synapse) with the current potential function,
	 *  thus forming a new potential function.
	 *  This is a heavily adapted merge-sort algorithm, so it should be quite
	 *  fast.
	 */
	public void mergeSynapseResponse(Object change) {
		Debug.output(Debug.LEVEL_INFO, "PLPotential.mergeSynapseResponse for " + label +
			" at time " + getTime() + "\n");
		/* this only works for potential changes that are generated from
		   PLSynapses */
		Debug.assertCond(change instanceof PLFunction,
			"PLPotential.mergeSynapseResponse",
			"change must be of type PLFunction");
		PLFunction newChange = (PLFunction) change;
		int i = 0, j = 0, k = 0;
		boolean changesActive = false, newChangeActive = false;
		double endTime;
		// first update the lastValue to the current simulation time
		/*FloatIntPair lastSeg = getFinishedSegments();
		lastValue = lastSeg.f;
		i = lastSeg.i;*/
		//lastTime = changes.gradientTimes[i];

		/* this is a hack for the moment: always create a new array that is
		   definitely large enough to hold the new potential function. it should
		   go away, fixe sized arrays should be used and the array size should
		   be determined correctly when constructing this object. */
		buffer = new PLFunction(numSegments + newChange.gradients.length);

		Debug.output(Debug.LEVEL_TRACE, "Old potential:" +
			changes.toString(numSegments) + "ending at " +
			changes.getEndTime(0, numSegments) + "\nSynapse response to be merged:\n" +
			newChange.toString() + "ending at " +
			newChange.getEndTime() + "\n");
		// deactivate the neuron since it will get reactivated anyways
		passivate();
		// now the real merge sort of the future change times
		while (i < numSegments || j < newChange.gradients.length) {
			/* In the time comparisons, some inaccuracy window is allowed to
			   "catch" points that are very close together and threat them as
			   one (with correct gradient handling of course). This solves the
			   following problem (occured once):
			   When there are 2 functions to be merged, one EPSP and one IPSP
			   and both have nearly the same gradient times, then the last
			   segment might not reach zero if the inaccuracy is not corrected.
			   Example: changes = { (-3.5, 3.3333333333333335),
									(2.0, 4.333333333333334),
									(0.0, 6.083333333333335) }
						newChange = { (14.0, 0.0), (-8.0, 1.0), (0.0, 2.75) }
						at 3.3333333333333335.
			   This resulted in the following buffer:
					{ (10.5, 3.3333333333333335), (-6.0, 4.333333333333334),
					  (2.0, 6.083333333333334), (0.0, 6.083333333333335) }
			   which does, when the last segment is cut away, not end at zero.
			*/
			if (j >= newChange.gradients.length || // no more new changes
			   (i < numSegments &&                 // but there are more old segments
			   (newChange.gradientTimes[j] + getTime()) -
				changes.gradientTimes[i] >
				GlobalConstants.DOUBLE_INACCURACY_WINDOW)) {
				/* take a segment from the old changes as the next segment
				   --> if the old changes function was not "active" before, it
				   is now. there is no need to set it inactive since it's last
				   segment has a gradient of 0 anyways (which can be added to
				   the other gradient without causing problems). */
				changesActive = true;
				buffer.gradients[k] = changes.gradients[i] +
					(newChangeActive ? newChange.gradients[j-1] : 0);
				buffer.gradientTimes[k] = changes.gradientTimes[i];
				i++;
			}
			else if (i >= numSegments ||            // no more old changes
				 (j < newChange.gradients.length && // but more new segments
				 changes.gradientTimes[i] -
				 (newChange.gradientTimes[j] + getTime()) >
				 GlobalConstants.DOUBLE_INACCURACY_WINDOW)) {
				newChangeActive = true;
				buffer.gradients[k] = newChange.gradients[j] +
					(changesActive ? changes.gradients[i-1] : 0);
				buffer.gradientTimes[k] = newChange.gradientTimes[j] + getTime();
				j++;
			}
			else {
				/* the following as to be true here:
				   1. (j < newChange.gradients.length) because first if did not
				   match
				   2. (i < numSegments) because second if did not match
				   3. (i >= numSegments) ||
					  (changes.gradientTimes[i] >= newChange.gradientTimes[j] + getTime())
					  because first if did not match
				   4. (j >= newChange.gradients.length) ||
					  (newChange.gradientTimes[j] + getTime() >= change.gradientTimes[i])
					  because second if did not match
				   5. from 2. and 3. follows:
					  (changes.gradientTimes[i] >= newChange.gradientTimes[j] + getTime())
				   6. from 1. and 4. follows:
					  (newChange.gradientTimes[j] + getTime() >= change.gradientTimes[i])
				   7. form 5. and 6. follows:
					  changes.gradientTimes[i] == newChange.gradientTimes[j] + getTime()
				*/
				changesActive = newChangeActive = true;
				buffer.gradients[k] = changes.gradients[i] +
					newChange.gradients[j];
				buffer.gradientTimes[k] = changes.gradientTimes[i];
				i++; j++;
			}
			k++;
		}

		// Go back all the last segments that had a gradient of 0.
		while (k > 0 && buffer.gradients[k-1] == 0)
			k--;
		/* Insert the zero gradient segment again. This procedure is only
		   needed because of calculation inaccuracies. In an ideal world, the
		   last segment (which is guaranteed to have a gradient of zero) would
		   being exactly when the freshly merged potential function reached 0.
		   But since we use floating point numbers, this step is needed to
		   achieve accurate computations. It can be omitted, but then
		   getFinishedSegments(true) would have to cope with segments that do
		   not exactly end at the current simulation time. The inaccuray window
		   seems to be a critical parameter, so I am omitting this crux here.

		   However, this can only be done if there is room (if there isn't, then
		   the buffer has 0 elements and we get away with it :-) ).
		*/
		if (k < buffer.gradients.length) {
			buffer.gradients[k] = 0;
			buffer.gradientTimes[k] = buffer.getEndTime(0, k);
			k++;
		}
		numSegments = k;

		// now the buffer contains the future change information
		PLFunction swap = buffer;
		buffer = changes;
		changes = swap;

		Debug.output(Debug.LEVEL_TRACE, "New potential:" +
			changes.toString(numSegments));

		/* recalculate the time when the potential again reaches the value 0
		   (the quiescent neuron state) and activate at that time to set it to 0
		*/
		//endTime = changes.getEndTime(lastValue, numSegments);
		endTime = changes.getEndTime(0, numSegments);
		if (! isActive() && endTime > 0) {
			Debug.output(Debug.LEVEL_INFO, "potential object scheduled for activation at time "
				+ endTime + "\n");
			activateAt(endTime);
		}

		// notify all listeners that this object has changed its internal state
		signalVariableChange();
	}

	/** This method is called by the simulation framework whenever a time event
	 *  occurs. It is used within PLPotential for clearing the potential
	 *  function when it has reached a value of zero at some time (this can be
	 *  the end of the function, but does not have to be).
	 */
	public void processSimEvent(SimEvent e) {
		// skip segments that are in the past (or end now)
		PLFunction.FloatIntPair ret = changes.getFinishedSegments(
			0.0, numSegments, getTime(), true);

		/* Be sure that we are either at the end of the function or that it is
		   currently in a segment with a value of zero. */
		Debug.assertCond(ret.i == numSegments ||
			Math.abs(ret.f) <= GlobalConstants.DOUBLE_INACCURACY_WINDOW ||
			(ret.i == numSegments-1 && Math.abs(ret.f +
			changes.gradients[ret.i] *
			(getTime() - changes.gradientTimes[ret.i])) <=
			GlobalConstants.DOUBLE_INACCURACY_WINDOW),
			"PLPotential.processSimEvent",
			"function is not zero when trying to perform cleanup: " +
			"ret.i=" + ret.i + ", ret.f=" + ret.f +
			", numSegments=" + numSegments);
		if (ret.i == numSegments || (ret.i == numSegments-1 &&
			Math.abs(ret.f + changes.gradients[ret.i] *
			(getTime() - changes.gradientTimes[ret.i])) <=
			GlobalConstants.DOUBLE_INACCURACY_WINDOW)) {
			/* this is a special case, because it is the last segment - just
			   clear everything */
			numSegments = 0;
		}
		else {
			// now cut away the finished segments
			// this should (some time) work with arrays of fixed length
			for (int i=0; i<numSegments - ret.i; i++) {
				changes.gradients[i] = changes.gradients[i + ret.i];
				changes.gradientTimes[i] = changes.gradientTimes[i + ret.i];
			}
			numSegments -= ret.i;
		}

		// notify all listeners that this object has changed its internal state
		signalVariableChange();
	}

	/** Returns the currently computed piecewiese linear function describing
	 *  the currently known future potential changes.
	 *  Note: The length of the returned array will be of maximum size whereas
	 *  only a smaller number of segments is valid within the function. Use
	 *  getNumSegments() to retrieve the real number of valid segements, do not
	 *  user getChanges().gradients.length !
	 *  @see #getNumSegments
	 */
	public PLFunction getChanges() {
		return changes;
	}

	/** Returns the number of piecewiese linear segments that are currently
	 *  valid in the changes function.
	 */
	public int getNumSegments() {
		return numSegments;
	}
}
