/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: Synapse.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.8 $
 *  Last changed at: $Date: 2002/05/16 20:43:26 $
 *  Last changed by: $Author: rene $
 *****************************************************************************/

package mosaic.sim.neuron;

import mosaic.sim.kernel.*;

/** This class represents synapses, which are the connections between neurons.
 *  When a synapse object receives a spike event, it will forward it to the
 *  connected "post-synaptic" neuron object, which will then query the synapse
 *  for the response parameters. These parameters define how the neuron
 *  potential changes because of the incoming spike.
 *  This class is declared abstract
 */
public abstract class Synapse extends BasicActiveModel implements NeuronInput {
	/** This interface is used for external learning algorithms that are
	 *  non-local to the synapse. The default implementation of
	 *  postNeuronFired() calls the LearningMethod, if it has been registered
	 *  previously (and postNeutonFired() was not overridden).
	 *  Obviuosly, since this abstract class does not have any synaptic
	 *  parameters, the LearningMethod will depend on the implementation of the
	 *  synapse.
	 *
	 *  @see #learningMethod
	 *  @see #postNeuronFired
	 */
	public interface LearningMethod {
		public void changeParameters(Synapse sender);
	}

	/** The post-synaptic neuron that should receive the forwarded spike events.
	 */
	protected Neuron receiver;

	/** The pre-synaptic object that sends spike events to this synapse.
	 */
	protected NeuronOutput sender;

	/** The time when the last spike event has been received by this synapse.
	 */
	protected double lastEventTime;

	/** The learning method to use whenever changing the synaptic parameters is
	 *  to be done (e.g. when postNeuronFired() is called).
	 *  Can be set with setLearningMethod or via the constructor.
	 *  @see #postNeuronFired
	 *  @see #setLearningMethod
	 */
	protected LearningMethod learningMethod;

	/** A synapse object needs a pre-synaptic NeuronOutput object and a
	 *  post-synaptic neuron. This constructor also registers the newly created
	 *  synapse object with the sender and receiver objects.
	 */
	public Synapse(NeuronOutput pre, Neuron post) {
		learningMethod = null;
		sender = pre;
		receiver = post;
		// If the given objects are valid, register the synapse.
		if (sender != null)
			sender.addOutput(this);
		if (receiver != null)
			receiver.addInputSynapse(this);
	}

	/** Does the same as Synapse(NeuronOutput pre, Neuron post), but also
	 *  sets the learning method.
	 *  @see #learningMethod
	 */
	public Synapse(NeuronOutput pre, Neuron post,
		   LearningMethod learningMethod) {
		this(pre, post);
		this.learningMethod = learningMethod;
	}

	/** Returns the post-synaptic neuron that receives the forwarded spike
	 *  events from this synapse.
	 */
	public Neuron getReceiver() {
		return receiver;
	}

	/** Sets the post-synaptic neuron that receives the forwarded spike
	 *  events from this synapse.
	 */
	public void setReceiver(Neuron receiver) {
		this.receiver = receiver;
	}

	/** Returns the pre-synaptic NeuronOutput object that sends spike events to
	 *  this synapse.
	 */
	public NeuronOutput getSender() {
		return sender;
	}

	/** Sets the pre-synaptic NeuronOutput object that sends spike events to
	 *  this synapse.
	 */
	public void setSender(NeuronOutput sender) {
		this.sender = sender;
	}

	/** Sets the learning method.
	 */
	public void setLearningMethod(LearningMethod learningMethod) {
		this.learningMethod = learningMethod;
	}

	/** This is the implementation of the NeuronInput interface. It simply
	 *  stores the current simulation time in lastEventTime and forwards the
	 *  spike event to the post-synaptic neuron.
	 */
	public void spikeEvent(NeuronOutput sender) {
		Debug.output(Debug.LEVEL_INFO, "Synpase.spikeEvent for " + label +
			" at time " + getTime() + "\n");
		lastEventTime = SimApplication.getCurrentSimApplication().getTime();
		receiver.spikeEventFrom(this);
		Debug.output(Debug.LEVEL_INFO, "spikeEvent forwarded successfully\n");
	}

	/** This function is called by the post-synaptic neuron to retrieve the
	 *  parameters which describe how the neuron potential changes because of
	 *  the receives spike event.
	 */
	public abstract Object getResponseParameters();

	/** The neuron calls this method whenever it fires a spike event to notify
	 *  its input synapses of its firing. This is important for learning.
	 *  This implementation calls the initialized learningMethod if it has been
	 *  set (either via the constructor or via setLearningMethod).
	 *
	 *  It can be overridden to implement local-only learning (using only the
	 *  information that is locally available to the synapse).
	 */
	public void postNeuronFired() {
		if (learningMethod != null)
		   learningMethod.changeParameters(this);
	}

	/** Returns the time when this synapse has received the last spike. */
	public double getLastEventTime() {
		return lastEventTime;
	}

	/** Returns a copy of the current object.
	 */
	public Object clone() {
		try {
			return super.clone();
		} catch (CloneNotSupportedException e) {
			Debug.assertCond(false, "Synapse.clone", "CloneNotSupportedException");
			return null;
		}
	}
}
