/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: RandomSpikeSource.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.6 $
 *  Last changed at: $Date: 2002/05/16 20:43:31 $
 *  Last changed by: $Author: rene $
 *****************************************************************************/

package mosaic.sim.neuron.netinputs;

import java.util.*;
import mosaic.sim.neuron.*;
import mosaic.sim.kernel.simevent.*;

/** This class implements a random spike source with a configurable maximum
 *  interval between two emitted spikes. After being initializes with init(),
 *  it just fires spike events at random intervals [0, maxTime[.
 *
 *  Therefore this input can act for generating events in a poisson process.
 */
public class RandomSpikeSource extends NeuronOutputSupport {
	/** The random number generator used for getting the fire times. */
	private static Random rand = null;
	/** The maximum time between two spike events. */
	private double maxTime;

	/** Creates a RandomSpikeSource with the given maxTime.
	 *  @param maxTime Specifies the maximum time between two emitted spikes.
	 *  @see #maxTime
	 */
	public RandomSpikeSource(double maxTime) {
		this.maxTime = maxTime;
	}

	/** Reset the spike source. This is needed by the simulation framework for
	 *  correct behaviour.
	 */
	public void reset() {
		super.reset();
		init();
	}

	/** Initialize the random number generator and start the spike firing. */
	public void init() {
		if (rand == null)
			rand = new Random();
		double nextTime = rand.nextDouble() * maxTime;
		System.out.println("RandomSpikeSource " + label +
			": first activation at time " + nextTime);
		activateIn(nextTime);
	}

	/** This method is called by the simulation framework whenever a time event
	 *  occurs. It is used for firing the spike events and scheduling the next
	 *  event.
	 */
	public void processSimEvent(SimEvent e) {
		fireSpikeEvent();
		double nextTime = rand.nextDouble() * maxTime;
		System.out.println("RandomSpikeSource " + label +
			": next activation in " + nextTime);
		activateIn(nextTime);
	}
}
