/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: RateCodedInputSingle.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.2 $
 *  Last changed at: $Date: 2002/05/16 20:43:31 $
 *  Last changed by: $Author: rene $
 *****************************************************************************/

package mosaic.sim.neuron.netinputs;

import java.util.*;
import mosaic.sim.kernel.*;
import mosaic.sim.kernel.simevent.*;
import mosaic.sim.neuron.*;

/** This class implements a single spike source using rate coding. It is a
 *  converter between analog values at its input and spike events at its
 *  output.
 *  The coding is done as follows: The method getRate and setRate can be used
 *  to set the firing rate of the current input, which continously generates
 *  spike events at the given interval. The rate is specified as
 *  1 / delay_between_spike .
 */
public class RateCodedInputSingle extends NeuronOutputSupport {
	/** This is the current firing rate, specified by 1 / delay_between_spikes .
	 *  If the rate is set to 0, then the input is disabled.
	 */
	private double rate;

	/** Constructs a new RateCodedInput object and initializes the rate with 0
	 *  (disables the input).
	 */
	public RateCodedInputSingle() {
		this(0.0);
	}

	/** Constructs a new RateCodedInput object.
	 */
	public RateCodedInputSingle(double rate) {
		this.rate = rate;
	}

	/** Reset the spike source. This is needed by the simulation framework for
	 *  correct behaviour.
	 */
	public void reset() {
		super.reset();
		init();
	}

	/** Initialize the random number generator and start the spike firing. */
	public void init() {
		if (rate != 0.0) {
			double nextTime = 1 / rate;
			System.out.println("RateCodedInput " + label +
				": first activation at time " + nextTime);
			activateIn(nextTime);
		}
	}

	/** This method is called by the simulation framework whenever a time event
	 *  occurs. It is used for firing the spike events and scheduling the next
	 *  event.
	 */
	public void processSimEvent(SimEvent e) {
		fireSpikeEvent();
		if (rate != 0.0) {
			double nextTime = 1 / rate;
			System.out.println("RateCodedInput " + label +
				": next activation in " + nextTime);
			activateIn(nextTime);
		}
	}

	/** Returns the currently used firing rate.
	 *  @see #rate
	 */
	public double getRate() {
		return rate;
	}

	/** Sets the currently used firing rate. It will get used as soon as the
	 *  spike event occurs, so if the rate was very low before setting a new
	 *  rate, it might take some time before the new rate gets used. Use the
	 *  method reset() to immediately send the next spike and therfore apply
	 *  the new rate.
	 *  reset() is called internally if the previously set rate was 0 (because
	 *  in this case, the new rate would never get active).
	 *
	 *  @see #rate
	 *  @see #reset
	 */
	public void setRate(double rate) {
		double oldRate = rate;
		this.rate = rate;
		if (oldRate == 0.0)
		   reset();
	}
}
