/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: OutputFilter.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.2 $
 *  Last changed at: $Date: 2002/04/18 14:39:03 $
 *  Last changed by: $Author: Rene $
 *****************************************************************************/

package mosaic.sim.neuron.netoutputs;

import mosaic.sim.neuron.*;

/** This class is a decorator for output listeners, enabling the filtering of
 *  output events:
 *  - first of all, it can pass only the first event and filter out all
 *    subsequent ones
 *  - secondly, events can be forwarded when the have reached an equilibrium,
 *    i.e. they do not change anymore within a given error range
 *
 *  The constructor intializes the filter. When forwardMultipleEvents is
 *  true, then all events will be forwarded (depending on the value of
 *  waitForEquilibrium). If it is false, only the first event (also depending on
 *  waitForEquilibrium) will be forwarded.
 *  When waitForEquilibrium is true, then an event will forwarded only when the
 *  difference to the previously received event is within a certain, given
 *  margin.
 *  E.g. when waitForEquilibrium and forwardMultipleEvents are both true,
 *  then the filter discards all events that change too much and forwards the
 *  first one that has a small difference (as given by maxSquareError). After
 *  this forwarding, the filter resets and waits for the next equilibrium.
 *
 *  The "error", i.e. the difference between two received events, is calculated
 *  as the squared difference between the vectors or matrices. When the error
 *  is below the given maxSquareError, then the state is regarded as being an
 *  equilibrium. Note that the error is divided by the number of elements in
 *  the vector or matrix for normalization.
 */
public class OutputFilter
	   implements VectorOutputListener, MatrixOutputListener {
	/** The object to which the filtered vector events will be forwarded. If
	 *  this is null, then the vector events will not be forwarded.
	 */
	protected VectorOutputListener nextVectorListener;
	/** The object to which the filtered matrix events will be forwarded. If
	 *  this is null, then the matrix events will not be forwarded.
	 */
	protected MatrixOutputListener nextMatrixListener;

	/** When true, then multiple events will be forwarded. When false, only a
	 *  single event will ever be forwarded. See the class description for
	 *  details.
	 */
	protected boolean forwardMultipleEvents;
	/** When true, then only those events will be forwarded that are only
	 *  marginally different from the previous. See the class description for
	 *  details.
	 */
	protected boolean waitForEquilibrium;
	/** The maximum square error for determining if the received event is in an
	 *  equilibrium state. See the class description for details.
	 */
	protected double maxSquareError;

	/** True when an event has already been forwarded. */
	protected boolean forwardedFirstEvent;
	/** The last vector that has been received - used for determining if an
	 *  equilibrium has been reached.
	 */
	protected double[] lastVector;
	/** The last matrix that has been received - used for determining if an
	 *  equilibrium has been reached.
	 */
	protected double[][] lastMatrix;

	/** Calculates the normalized (divided by the vector length) square
	 *  difference between the given vectors.
	 */
	protected double calcSquareError(double[] lastVector, double[] vector) {
		double error = 0.0;
		Debug.assertCond(lastVector.length == vector.length,
					 "OutputFilter.calcSquareError",
					 "The input vectors must have equal length.");

		for (int i=0; i<lastVector.length; i++) {
			error += Math.pow((lastVector[i] - vector[i]), 2);
		}
		error /= lastVector.length;
		return error;
	}

	/** Calculates the normalized (divided by number of elements in the matrix
	 *  length) square difference between the given matrices.
	 */
	protected double calcSquareError(double[][] lastMatrx, double[][] matrix) {
		double error = 0.0;
		Debug.assertCond(lastMatrix.length == matrix.length,
					 "OutputFilter.calcSquareError",
					 "The input matrices must have equal lengths.");

		for (int i=0; i<lastMatrix.length; i++) {
			error += calcSquareError(lastMatrix[i], matrix[i]);
		}
		error /= lastMatrix.length;
		return error;
	}

	/** Constructs an OuptutFilter object and initializes the mamber variables
	 *  according to the given parameters.
	 *
	 *  @param nextVectorListener See member variable nextVectorListener
	 *  @param nextMatrixListener See member variable nextMatrixListener
	 *  @param forwardMultipleEvents See member variable forwardMultipleEvents
	 *  @param waitForEquilibrium See member variable waitForEquilibrium
	 *  @param maxSquareError See member variable maxSquareError
	 */
	public OutputFilter(VectorOutputListener nextVectorListener,
		   MatrixOutputListener nextMatrixListener,
		   boolean forwardMultipleEvents,
		   boolean waitForEquilibrium, double maxSquareError) {
		Debug.assertCond(!waitForEquilibrium || maxSquareError > 0.0,
			"OutputFilter.OutputFiler",
			"maxSquareError must be > 0.0 when waitForEquilibrium is true");

		this.nextVectorListener = nextVectorListener;
		this.nextMatrixListener = nextMatrixListener;
		this.forwardMultipleEvents = forwardMultipleEvents;
		this.waitForEquilibrium = waitForEquilibrium;
		this.maxSquareError = maxSquareError;

		forwardedFirstEvent = false;
		lastVector = null;
		lastMatrix = null;
	}

	/** Implementation of VectorOutputListener.outputEvent. */
	public void outputEvent(double[] vector) {
		// only do something if we can forward the filtered event
		if (nextVectorListener != null) {
			// check if we should only forward the first event
			if (forwardMultipleEvents || !forwardedFirstEvent) {
				// now the equilibrium check
				if (!waitForEquilibrium || ( lastVector != null &&
				   calcSquareError(lastVector, vector) < maxSquareError)) {
					nextVectorListener.outputEvent(vector);
					forwardedFirstEvent = true;
					lastVector = null;
				}
				else if (waitForEquilibrium)
					 lastVector = vector;
			}
		}
	}

	/** Implementation of MatrixOutputListener.outputEvent. */
	public void outputEvent(double[][] matrix) {
		// only do something if we can forward the filtered event
		if (nextMatrixListener != null) {
			// check if we should only forward the first event
			if (forwardMultipleEvents || !forwardedFirstEvent) {
				// now the equilibrium check
				if (!waitForEquilibrium || ( lastMatrix != null &&
				   calcSquareError(lastMatrix, matrix) < maxSquareError)) {
					nextMatrixListener.outputEvent(matrix);
					forwardedFirstEvent = true;
					lastMatrix = null;
				}
				else if (waitForEquilibrium)
					 lastMatrix = matrix;
			}
		}
	}
}
