/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: TemporalCodedOutput.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.9 $
 *  Last changed at: $Date: 2002/05/16 20:43:33 $
 *  Last changed by: $Author: rene $
 *****************************************************************************/

package mosaic.sim.neuron.netoutputs;

import java.util.*;
import mosaic.sim.kernel.*;
import mosaic.sim.kernel.simevent.*;
import mosaic.sim.neuron.*;

/** This class implements a system output class. It is a converter between
 *  spike events at its inputs and analog vectors at its output. The spike
 *  events have to be coded temporally, e.g. from a TemporalCodedInput. For
 *  a description of temporal coding, please look at TemporalCodedInput.
 *  @see mosaic.sim.neuron.netinputs.TemporalCodedInput
 */
public class TemporalCodedOutput extends BasicActiveModel {
	/** This inner class implements the inputs of TemporalCodedOutput, i.e.
	 *  the objects receiving the spike events from the system.
	 */
	protected class Input implements NeuronInput {
		/** Set to true when this input has already received a spike event in
		 *  this output cycle. Only one spike event can be received within one
		 *  cycle, a second one would be a coding violation.
		 */
		protected boolean receivedSpike = false;
		/** The simulation time when the spike event was reveived by this
		 *  input.
		 */
		protected double receivedTime = 0;

		/** This function is the implementation of NeuronInput.spikeEvent and
		 *  is called when a spike event is received.
		 */
		public void spikeEvent(NeuronOutput sender) {
			if (receivedSpike) {
				// this is a temporal code violation
				//throw new CodingException("Invalid (second) spike received.");
				Debug.assertCond(false, "TemporalCodedOutput.Input.spikeEvent",
					"Received second (invalid) spike.");
			}
			receivedSpike = true;
			receivedTime = SimApplication.getCurrentSimApplication().getTime();
			System.out.println("TemporalCodedOutput.Input.spikeEvent: " +
				"received spike at time " + receivedTime);
			spikeEventOccured();
		}

		/** Returns true if this input has already received a spike.
		 *  @see #receivedSpike
		 */
		public boolean hasReceivedSpike() {
			return receivedSpike;
		}

		/** Resets the input after the output cycle is completed. Must be
		 *  called before another spike can be received.
		 */
		public void reset() {
			receivedSpike = false;
		}

		/** Returns the time when the input has received its spike.
		 *  @see #receivedTime
		 */
		public double getReceivedTime() {
			return receivedTime;
		}
	}

	/** The listeners which should receive outputEvents sent by this class. */
	protected ArrayList listeners = null;
	/** The inputs which receive the spike series. */
	protected ArrayList inputs = null;
	/** An optional reference input representing the analog value "0". */
	protected Input reference = null;
	/** Multiplicator for computing the analog value depending on the delay
	 *  between the received spikes.
	 *  @see mosaic.sim.neuron.netinputs.TemporalCodedInput#multiplicator
	 */
	protected double multiplicator;
	/** This is the maximum time that an input cycle is allow to take. After
	 *  this time has passed (after receiving the first spike event in the input
	 *  cycle), all inputs that have not received any spike are set to a value
	 *  NEGATIVE_INFINITY.
	 */
	protected double maxCycleTime;
	/** This boolean variable is set to true whenever an output cycle is
	 *  currently active.
	 */
	protected boolean cycleActive = false;

	/** This method is called by the input objects whenever a spike event has
	 *  been received. It checks if the output cycle is complete and sends the
	 *  output events if it is.
	 */
	protected void spikeEventOccured() {
		/* first check if all spikes have been received yet
		   If a reference spike is used, the cycle is completed when it
		   receives the reference spike. If not, then all inputs have to
		   receive a spike to complete the cycle. */
		System.out.print("TemporalCodedOuput.spikeEventOccured: ");
		boolean complete = true;
		if (reference == null) {
			for (int i=0; complete && i<inputs.size(); i++)
				if (! ((Input) inputs.get(i)).hasReceivedSpike())
					complete = false;
			System.out.println("not using reference, complete=" + complete);
		}
		else {
			complete = reference.hasReceivedSpike();
			System.out.println("using reference, hasReceivedSpike=" + complete);
		}

		if (!cycleActive) {
			System.out.println("Output cycle started.");
			/* This is the first spike that is starting the output cycle -
			   set the timer for the maximum time. Note that this will
			   automatically get reset when the cycle is completed (see
			   the call to activateIn() below this one). */
			activateIn(maxCycleTime);
			cycleActive = true;
		}

		if (complete) {
			/* This causes the activation of this class as soon as the currently
			   active event (which caused the reception of the spike event and
			   therefore the start of this method) has finished its handler,
			   but still at the current simulation time. */
			activateIn(0);
		}
	}

	/** Creates a TemporalCodedOuput object and initializes the input objects.
	 *  @param num The number of inputs. It is equal to the length of the
	 *             analog vectors sent in the output events.
	 *  @param multiplicator The multiplicator.
	 *  @param useReferenceInput When set to true, the class uses a reference
	 *                           input representing the analog input value
	 *                           "0".
	 */
	public TemporalCodedOutput(int num, double multiplicator, boolean useReferenceInput,
		   double maxCycleTime) {
		listeners = new ArrayList();
		inputs = new ArrayList(num);
		for (int i=0; i<num; i++) {
			Input input = new Input();
			inputs.add(i, input);
			//add(input);
		}
		if (useReferenceInput)
			reference = new Input();
		this.multiplicator = multiplicator;
		this.maxCycleTime = maxCycleTime;
		cycleActive = false;
	}

	/** This method is called by the simulation framework whenever a time event
	 *  for this object occurs. The only purpose in this class is to decouple
	 *  the receiption of the last spike event (completing the output cycle)
	 *  and sending of the output event. This way, the event handlers of the
	 *  classes which caused the spike events to be sent in the first place
	 *  (e.g. a TemporalCodedInput object) can complete before this output
	 *  event is sent.
	 *  This was necessary to get a test example working, but it seems a lot
	 *  cleaner anyway. The output event is sent at the same simulation time
	 *  (because the class activates itself upon completion of the output
	 *  cycle with activateIn(0) ), but not during the executing of other time
	 *  event, but after them.
	 */
	public void processSimEvent(SimEvent e) {
		System.out.println("Output cycle ending");
		double outVector[] = new double[inputs.size()];

		for (int i=0; i<inputs.size(); i++) {
			Input inp = (Input) inputs.get(i);
			if (inp.hasReceivedSpike()) {
				outVector[i] = (getTime() - inp.getReceivedTime()) / multiplicator;
				inp.reset();
			}
			else
				/* If the corresponding input has not received a spike,
				   then this is indicated by the special value
				   NEGATIVE_INFINITY. */
				outVector[i] = Double.NEGATIVE_INFINITY;
		}

		// if we use a reference, also reset this one...
		if (reference != null)
		   reference.reset();

		// now the cycle has ended
		cycleActive = false;

		// send the real output event now
		for (int i=0; i<listeners.size(); i++)
			((VectorOutputListener) listeners.get(i)).outputEvent(outVector);
	}

	/** Returns the input object with the given index.
	 *  @param index The number of the input object to return.
	 *  @see #inputs
	 */
	public NeuronInput getInput(int index) {
		if (index >= 0 && index < inputs.size())
			return (Input) inputs.get(index);
		else
			throw new ArrayIndexOutOfBoundsException();
	}

	/** Returns the reference input. */
	public NeuronInput getReferenceInput() {
		if (reference != null)
			return reference;
		else
			throw new NullPointerException();
	}

	/** Adds an output event listener.
	 *  @param l The listener that should be added to the list of listeners
	 *           receiving output events by this class.
	 *  @see #listeners
	 */
	public void addVectorOutputListener(VectorOutputListener l) {
		listeners.add(l);
	}

	/** Removes an output event listener.
	 *  @param l The listener that should be removed from the list of listeners
	 *           receiving output events by this class.
	 *  @see #listeners
	 */
	public boolean removeOutputEventListener(VectorOutputListener l) {
		return listeners.remove(l);
	}
}
