/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: Hopfield.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.8 $
 *  Last changed at: $Date: 2002/04/24 19:44:59 $
 *  Last changed by: $Author: Rene $
 *****************************************************************************/

package mosaic.sim.neuron.simulations;

import mosaic.sim.*;
import mosaic.sim.kernel.*;
import mosaic.sim.neuron.*;
import mosaic.sim.neuron.netinputs.*;
import mosaic.sim.neuron.netoutputs.*;
import mosaic.sim.neuron.visualisation.*;
import java.awt.*;
import javax.swing.*;

/* */
public class Hopfield extends SimApplication {

	public Hopfield() {
		SimpleMainFrame frame = new SimpleMainFrame(this);

// ========> create your panel here <==========================================//
		HopfieldPanel p=new HopfieldPanel();
		frame.setModelPanel(p);
// ============================================================================//

		frame.pack();
		frame.setVisible(true);
		frame.setTitle("Hopfield SNN simulation");
		reset();
	}

	public static void main(String[] args) {
		Hopfield app = new Hopfield();
	}

}

class HopfieldPanel extends ModelPanel {
	public static final int NUMBER_NEURONS = 10;
	public static final double RISING_GRADIENT = 1.0;
	public static final double RISING_TIME = 2.0;
	public static final double THRESHOLD = 1.5;
	public static final double[] SYNAPSE_DELAYS = { 10.0 };
	public static final double REFRACTION_TIME = 5.0;

	public static final double[][] ATTRACTOR_SET = {
		{ 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0 },
		{ 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0 },
		{ 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0 },
		{ 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0 } };

	public static final double[] INPUT_VECTOR =
		/*{ 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };*/
		{ 0.7, 0.3, 0.2, Double.NEGATIVE_INFINITY,
			Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY,
			Double.NEGATIVE_INFINITY, 0.1, 0.3 };
		/*{ Double.NEGATIVE_INFINITY, 1.0, Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY,
			Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY,
			Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY };*/

	public static final boolean SHOW_NEURON_DETAILS = false;

	Neuron[] neurons = new Neuron[NUMBER_NEURONS];
	TemporalCodedInput input;
	TemporalCodedOutput output;

	ViewSynchronizer sync = new ViewSynchronizer();
	ScrollingView[] views;

	PLFunction synapseResponse;

	public HopfieldPanel() {
		try {
			// the parameter T
			double T = /*SYNAPSE_DELAYS[0] +*/ THRESHOLD / RISING_GRADIENT;
			input = new TemporalCodedInput(NUMBER_NEURONS, T, false);
			output = new TemporalCodedOutput(NUMBER_NEURONS, T, false, T);
			// the weight values
			double weights[][] = new double[NUMBER_NEURONS][NUMBER_NEURONS];
			for (int i=0; i<NUMBER_NEURONS; i++) {
				for (int j=0; j<NUMBER_NEURONS; j++) {
					weights[i][j] = 0.0;
					for (int k=0; k<ATTRACTOR_SET.length; k++) {
						weights[i][j] +=
							ATTRACTOR_SET[k][i] * ATTRACTOR_SET[k][j];
					}
				}
			}
			/* normalize the weigths to be in the range of -1.0 ... 1.0
			   and so that they sum up to 0 */
			for (int i=0; i<NUMBER_NEURONS; i++) {
				double linesum = 0;
				for (int j=0; j<NUMBER_NEURONS; j++) {
					linesum += weights[i][j];
				}
				/* when all weights are zero, we can not divide through a
				   linesmum of zero... */
				if (linesum != 0.0) {
					for (int j=0; j<NUMBER_NEURONS; j++) {
						weights[i][j] = -linesum/NUMBER_NEURONS + weights[i][j] * 2 / linesum;
					}
				}
			}
			System.out.println("Weight table: ");
			for (int i=0; i<NUMBER_NEURONS; i++) {
				for (int j=0; j<NUMBER_NEURONS; j++) {
					System.out.print(weights[i][j] + " ");
				}
				System.out.println();
			}

			// some global settings - thresholds
			PLFunction tr = new PLFunction(0 /*2*/);
			/*tr.gradients[0] = -5;
			tr.gradientTimes[0] = 0;
			tr.gradients[1] = -1;
			tr.gradientTimes[1] = 0.5;*/
			ConstantPLThreshold.setAbsoluteRefractionTime((float) REFRACTION_TIME);
			ConstantPLThreshold.setQuiescentValue((float) THRESHOLD);
			ConstantPLThreshold.setRelativeRefractionStartValue((float) THRESHOLD + 5);
			ConstantPLThreshold.setRelativeRefraction(tr);
			ConstantPLThreshold.calculateRelativeRefractionTime();
			// synaptic responses
			synapseResponse = new PLFunction(3);
			synapseResponse.gradients[0] = RISING_GRADIENT;
			synapseResponse.gradientTimes[0] = 0;
			synapseResponse.gradients[1] = - RISING_GRADIENT;
			synapseResponse.gradientTimes[1] = RISING_TIME;
			synapseResponse.gradients[2] = 0;
			synapseResponse.gradientTimes[2] = synapseResponse.getEndTime(0, 2);
			System.out.println("Synapse response function set to:");
			System.out.print(synapseResponse.toString());
			ConstantPLSynapse.setResponse(synapseResponse);

			// create and connect the neurons
			for (int i=0; i<NUMBER_NEURONS; i++) {
				neurons[i] = new Neuron();
				neurons[i].setLabel("Neuron" + i);
				neurons[i].setName("Neuron" + i);

				// don't forget to create the potentials and thresholds
				PLPotential p = new PLPotential(1);
				p.setName("potential " + i);
				neurons[i].setPotential(p);
				ConstantPLThreshold t = new ConstantPLThreshold();
				t.setName("threshold " + i);
				neurons[i].setThreshold(t);
			}
			// create the lateral connections
			for (int i=0; i<NUMBER_NEURONS; i++) {
				for (int j=0; j<NUMBER_NEURONS; j++) {
					if (i != j) {
						for (int k=0; k<SYNAPSE_DELAYS.length; k++) {
							ConstantPLSynapse s = new ConstantPLSynapse(
								neurons[i], neurons[j],
								(float) weights[i][j], (float) SYNAPSE_DELAYS[k]);
						}
					}
				}
			}
			// and the input connections
			for (int i=0; i<NUMBER_NEURONS; i++) {
				new ConstantPLSynapse(input.getOutput(i), neurons[i]);
			}
			// finally the output connections
			for (int i=0; i<NUMBER_NEURONS; i++) {
				neurons[i].addOutput(output.getInput(i));
			}

			// and use the output values
			output.addVectorOutputListener(
				new VectorOutputListener() {
					public void outputEvent(double vector[]) {
						System.out.print("Output event: ");
						for (int i=0; i<vector.length; i++)
							System.out.print(vector[i] + " ");
						System.out.println();
					}
				} );

			// build up the layout
			setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
			add(input);
			views = new ScrollingView[NUMBER_NEURONS * 2 +
				  (SHOW_NEURON_DETAILS ? NUMBER_NEURONS : 0)];
			for (int i=0; i<NUMBER_NEURONS; i++) {
				add(neurons[i]);
				views[i] = new SpikeView("Input view " + i);
				add(views[i]);
				input.getOutput(i).addOutput((SpikeView) views[i]);

				views[NUMBER_NEURONS+i] = new SpikeView("Neuron view " + i);
				add(views[NUMBER_NEURONS+i]);
				neurons[i].addOutput((SpikeView) views[NUMBER_NEURONS+i]);

				if (SHOW_NEURON_DETAILS) {
					views[NUMBER_NEURONS*2+i] = new PLNeuronInternalView(
						"Neuron " + i, true, true);
					add(views[NUMBER_NEURONS*2+i]);
					neurons[i].getPotential().addVariableChangeListener(
						(PLNeuronInternalView) views[NUMBER_NEURONS*2+i]);
					neurons[i].getThreshold().addVariableChangeListener(
						(PLNeuronInternalView) views[NUMBER_NEURONS*2+i]);
				}
			}

			for (int i=0; i<views.length; i++) {
				views[i].setName("v" + (i+1));
				views[i].setTMax(50);
				views[i].setCurrentTimePosition(0.5);
				sync.addView(views[i]);
				this.add(views[i]);
				if (views[i] instanceof PLNeuronInternalView) {
				   views[i].setYMin(-1);
				   views[i].setYMax(15);
				}
			}
		}
		catch(Exception e) {
			e.printStackTrace();
		}
	}

	public void init() {
		// start the simulation by giving the input
		try {
			input.enterVector(INPUT_VECTOR);
		}
		catch (CodingException e) {
			e.printStackTrace();
		}
	}

	public void reset() {
		super.reset();
		init();
	}
}
