/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: SOM.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.4 $
 *  Last changed at: $Date: 2002/04/24 20:10:41 $
 *  Last changed by: $Author: Rene $
 *****************************************************************************/

package mosaic.sim.neuron.simulations;

import mosaic.sim.*;
import mosaic.sim.kernel.*;
import mosaic.sim.neuron.*;
import mosaic.sim.kernel.simevent.*;
import mosaic.sim.neuron.netinputs.*;
import mosaic.sim.neuron.netoutputs.*;
import mosaic.sim.neuron.visualisation.*;
import java.awt.*;
import java.io.*;
import java.util.*;
import javax.swing.*;

public class SOM extends SimApplication {
	public SOM() {
		SimpleMainFrame frame = new SimpleMainFrame(this);

// ========> create your panel here <==========================================//
		SOMPanel p=new SOMPanel();
		frame.setModelPanel(p);
// ============================================================================//

		frame.pack();
		frame.setVisible(true);
		frame.setTitle("SOM SNN simulation");
		reset();
	}

	public static void main(String[] args) {
		SOM app = new SOM();
	}

}

/** Note that we do NOT use weight normalization after each learning cycle
 *  here - this is closer to the biological model.
 */
class SOMLearningMethod implements Synapse.LearningMethod {
	protected SOMPanel panel;
	protected HashSet synapsesDoneInCycle;

	public SOMLearningMethod(SOMPanel panel) {
		this.panel = panel;
		synapsesDoneInCycle = new HashSet();
	}

	public void changeParameters(Synapse sender) {
		PLSynapse syn = (PLSynapse) sender;

		if (synapsesDoneInCycle.contains(sender)) {
			System.out.println("Synapse " + sender.getName() + " already done in this cycle");
			return;
		}

		synapsesDoneInCycle.add(sender);

		System.out.print("Computing delta w for Synapse " + sender.getName() + " : ");
		double deltaw = panel.learningRate *
			   (panel.next_Tout - SimApplication.getCurrentSimApplication().getTime()) /
			   panel.next_Tout *
			   /* the next two lines are equal to s_i (the input value of the
				  i-th neuron */
			   ((panel.input.getLastZeroTime() - syn.getLastEventTime()) /
				panel.input.getMultiplicator()
				- syn.getWeight());
		System.out.println(deltaw + ", old = " + syn.getWeight() + ", new = " + (syn.getWeight() + deltaw));
		syn.setWeight(syn.getWeight() + (float) deltaw);
	}

	public void startNewCycle() {
		synapsesDoneInCycle.clear();
	}
}

/** The construction in here is exactly as chapter 11 of the PHD Thesis of
 *  Berthold Ruf describes.
 */
class SOMPanel extends ModelPanel implements TimeEventListener {
	public static final String FILENAME_TRAININGSET = "SOM_training_set.dat";
	public static final String FILENAME_COMPUTEDWEIGHTS = "SOM_computed_weights.dat";

	protected mosaic.sim.kernel.simevent.Timer timer;
	protected Random rand;
	protected SOMLearningMethod brain;

	public static final int NUMBER_INPUTS = 10;
	public static final int NUMBER_NEURONS = 10;
	public static final double RISING_GRADIENT = 1.0;
	public static final double RISING_TIME = 2.0;
	public static final double THRESHOLD = 1.5;
	public static final double[] SYNAPSE_DELAYS = { 10.0 };
	public static final double REFRACTION_TIME = 5.0;

	public static final int LEARNING_CYCLES = 4000;
	public static final float LEARNINGRATE_START = 0.1F;
	public static final float LEARNINGRATE_FACTOR = 0.995F;
	public static final float NEIGHBOURHOOD_DECREASE = 0.0005F;
	protected int currentLearningCycle;

	/** set sufficiently large so that every competitive neuron actually fires */
	public static final double LAMBDA = 3.0;

	/** calculated from the above constants - the length of an input / output
	 *  interval
	 */
	protected double T;
	protected double learningRate;
	protected double next_Tout;

	/*public static final double[][] TRAINING_SET = {
		{ 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0 },
		{ 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0 },
		{ 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0 },
		{ 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0 } };*/
	public static final int TRAINING_SET_SIZE = 10;
	protected double[][] TRAINING_SET;

	public static final boolean SHOW_NEURON_DETAILS = false;

	Neuron[] neurons = new Neuron[NUMBER_NEURONS];
	TemporalCodedInput input;
	//TemporalCodedOutput output;

	ViewSynchronizer sync = new ViewSynchronizer();
	ScrollingView[] views;

	PLFunction synapseResponse;

	public SOMPanel() {
		try {
			timer = new mosaic.sim.kernel.simevent.Timer();
			timer.addTimeEventListener(this);

			brain = new SOMLearningMethod(this);

			learningRate = 1.0;

			Random rand = new Random();
			// first of all, before doing anything else, try to read the training set
			TRAINING_SET = null;
			try {
				FileInputStream f = new FileInputStream(FILENAME_TRAININGSET);
				ObjectInputStream s = new ObjectInputStream(f);
				TRAINING_SET = (double[][]) s.readObject();
				System.out.println("Training set successfully read from old file");
			}
			catch (Exception e) {
				System.err.println("Unable to load file " + FILENAME_TRAININGSET);
				System.err.println(e.getMessage());
				TRAINING_SET = new double[TRAINING_SET_SIZE][NUMBER_INPUTS];
				for (int i=0; i<TRAINING_SET_SIZE; i++)
					for (int j=0; j<NUMBER_INPUTS; j++)
						TRAINING_SET[i][j] = 0.1 + rand.nextDouble() * 0.8;
				// and write to the file so it can be used next time
				try {
					FileOutputStream f = new FileOutputStream(FILENAME_TRAININGSET);
					ObjectOutputStream s = new ObjectOutputStream(f);
					s.writeObject(TRAINING_SET);
					s.flush();
					f.flush();
					f.close();
					System.out.println("Training set successfully written to new file");
				}
				catch (Exception e2) {
					System.err.println("Aie! Fix this first!");
					System.err.println(e2.getMessage());
				}
			}
			System.out.println("Training set is now:");
			for (int i=0; i<TRAINING_SET.length; i++) {
				for (int j=0; j<NUMBER_INPUTS; j++)
					System.out.print(TRAINING_SET[i][j] + " ");
				System.out.println();
			}

			// *************** caclulations as specified in the paper ******
			// input weights
			/* this is done as follows: for each neuron j use the mean of the
			   training set plus some random value in the interval [-0.1, 0.1] */
			// +1 because a reference input is used
			double in_w[][] = new double[NUMBER_INPUTS+1][NUMBER_NEURONS];
			double mean_in[] = new double[NUMBER_INPUTS];
			for (int i=0; i<NUMBER_INPUTS; i++) {
				mean_in[i] = 0;
				for (int j=0; j<TRAINING_SET.length; j++)
					mean_in[i] += (TRAINING_SET[j][i] / TRAINING_SET.length);
			}
			// also normalize the weight vector and set the zero element
			for (int j=0; j<NUMBER_NEURONS; j++) {
				double linesum = 0.0;
				for (int i=0; i<NUMBER_INPUTS; i++) {
					in_w[i+1][j] = mean_in[i] - 0.1 + 0.2 * rand.nextFloat();
					linesum += in_w[i+1][j];
				}
				for (int i=0; i<NUMBER_INPUTS; i++)
					in_w[i+1][j] /= linesum;
				in_w[0][j] = LAMBDA - 1.0;
			}
			/* now for every weight vector in_w[*][j] (for every neuron j), the
			   sum of its elements is equal to LAMBA */

			// lateral weights
			double l_w[][] = new double[NUMBER_NEURONS][NUMBER_NEURONS];
			/* for computing the lateral weights, the neighberhood matrix is used
			   is is just defined as: m[k][j] = abs(k-j)
			   mmax is the largest element of the matrix and therefore equal to
			   NUMBER_NEURONS-1 */
			for (int i=0; i<NUMBER_NEURONS; i++)
				for (int j=0; j<NUMBER_NEURONS; j++)
					l_w[i][j] = 3 * (0.25 - Math.abs(i-j) / (double) (NUMBER_NEURONS-1));

			// the parameter T is the length of an input cycle
			T = SYNAPSE_DELAYS[0] + 10 * THRESHOLD / RISING_GRADIENT;
			input = new TemporalCodedInput(NUMBER_INPUTS, 1.0, true);
			//output = new TemporalCodedOutput(NUMBER_NEURONS, T, false, T);


			// *************** from here downwards are simulation specifics

			// some global settings - thresholds
			PLFunction tr = new PLFunction(0 /*2*/);
			/*tr.gradients[0] = -5;
			tr.gradientTimes[0] = 0;
			tr.gradients[1] = -1;
			tr.gradientTimes[1] = 0.5;*/
			ConstantPLThreshold.setAbsoluteRefractionTime((float) REFRACTION_TIME);
			ConstantPLThreshold.setQuiescentValue((float) THRESHOLD);
			ConstantPLThreshold.setRelativeRefractionStartValue((float) THRESHOLD + 5);
			ConstantPLThreshold.setRelativeRefraction(tr);
			ConstantPLThreshold.calculateRelativeRefractionTime();
			// synaptic responses
			synapseResponse = new PLFunction(3);
			synapseResponse.gradients[0] = RISING_GRADIENT;
			synapseResponse.gradientTimes[0] = 0;
			synapseResponse.gradients[1] = - RISING_GRADIENT;
			synapseResponse.gradientTimes[1] = RISING_TIME;
			synapseResponse.gradients[2] = 0;
			synapseResponse.gradientTimes[2] = synapseResponse.getEndTime(0, 2);
			System.out.println("Synapse response function set to:");
			System.out.print(synapseResponse.toString());
			ConstantPLSynapse.setResponse(synapseResponse);

			// create and connect the neurons
			for (int i=0; i<NUMBER_NEURONS; i++) {
				neurons[i] = new Neuron();
				neurons[i].setLabel("Neuron" + i);
				neurons[i].setName("Neuron" + i);

				// don't forget to create the potentials and thresholds
				PLPotential p = new PLPotential(1);
				p.setName("potential " + i);
				neurons[i].setPotential(p);
				ConstantPLThreshold t = new ConstantPLThreshold();
				t.setName("threshold " + i);
				neurons[i].setThreshold(t);
			}
			// create the input connections
			System.out.println("Initialized input weights:");
			for (int j=0; j<NUMBER_NEURONS; j++) {
				System.out.print(in_w[0][j] + " ");
				new ConstantPLSynapse(input.getReferenceOutput(), neurons[j],
					(float) in_w[0][j], 0.0F, brain);
				for (int i=0; i<NUMBER_INPUTS; i++) {
					System.out.print(in_w[i+1][j] + " ");
					new ConstantPLSynapse(input.getOutput(i), neurons[j],
						(float) in_w[i+1][j], 0.0F, brain);
				}
				System.out.println();
			}
			// and the lateral connections
			System.out.println("Initialized lateral weights:");
			for (int i=0; i<NUMBER_NEURONS; i++) {
				for (int j=0; j<NUMBER_NEURONS; j++) {
					if (i!=j) {
						System.out.print(l_w[i][j] + " ");
						/* definitely no learning method here - the weights are
						   simply decreased as learning proceeds */
						new ConstantPLSynapse(neurons[i], neurons[j], (float) l_w[i][j]);
					}
					else
						System.out.print("0 ");
				}
				System.out.println();
			}
			// finally the output connections
			/*for (int i=0; i<NUMBER_NEURONS; i++) {
				neurons[i].addOutput(output.getInput(i));
			}

			// and use the output values
			output.addVectorOutputListener(
				new VectorOutputListener() {
					public void outputEvent(double vector[]) {
						System.out.print("Output event: ");
						for (int i=0; i<vector.length; i++)
							System.out.print(vector[i] + " ");
						System.out.println();
					}
				} );*/

			// build up the layout
			setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
			views = new ScrollingView[NUMBER_NEURONS * 2 + 1 +
				  (SHOW_NEURON_DETAILS ? NUMBER_NEURONS : 0)];
			views[NUMBER_NEURONS] = new SpikeView("Input view " + "reference", Color.cyan);
			input.getReferenceOutput().addOutput((SpikeView) views[NUMBER_NEURONS]);
			for (int i=0; i<NUMBER_NEURONS; i++) {
				views[i] = new SpikeView("Input view " + i, Color.blue);
				add(views[i]);
				input.getOutput(i).addOutput((SpikeView) views[i]);

				views[NUMBER_NEURONS+1+i] = new SpikeView("Neuron view " + i, Color.magenta);
				add(views[NUMBER_NEURONS+1+i]);
				neurons[i].addOutput((SpikeView) views[NUMBER_NEURONS+1+i]);

				if (SHOW_NEURON_DETAILS) {
					views[NUMBER_NEURONS*2+1+i] = new PLNeuronInternalView(
						"Neuron " + i, true, true);
					add(views[NUMBER_NEURONS*2+1+i]);
					neurons[i].getPotential().addVariableChangeListener(
						(PLNeuronInternalView) views[NUMBER_NEURONS*2+1+i]);
					neurons[i].getThreshold().addVariableChangeListener(
						(PLNeuronInternalView) views[NUMBER_NEURONS*2+1+i]);
				}
			}

			for (int i=0; i<views.length; i++) {
				views[i].setName("v" + (i+1));
				views[i].setTMax(50);
				views[i].setCurrentTimePosition(0.5);
				sync.addView(views[i]);
				this.add(views[i]);
				if (views[i] instanceof PLNeuronInternalView) {
				   views[i].setYMin(-1);
				   views[i].setYMax(15);
				}
			}
		}
		catch(Exception e) {
			e.printStackTrace();
		}
	}

	public void init() {
		// start the simulation by giving the first input
		rand = new Random();
		currentLearningCycle = 0;
		processSimEvent(null);
	}

	public void reset() {
		super.reset();
		init();
	}

	public void processSimEvent(SimEvent evt) {
		System.out.println("*** Starting training session - doing reset ***");
		/* every time we start a new training session, reset all competitive
		   neurons */
		for (int i=0; i<NUMBER_NEURONS; i++)
			neurons[i].reset();
		// and initialize the trainer
		brain.startNewCycle();

		// slowly decrease the learning rate
		learningRate *= LEARNINGRATE_FACTOR;

		// and the neighborhood
		for (int i=0; i<NUMBER_NEURONS; i++) {
			Synapse[] synapses = neurons[i].getInputSynapses();
			for (int j=0; j<synapses.length; j++) {
				PLSynapse syn = (PLSynapse) synapses[j];
				syn.setWeight(syn.getWeight() - NEIGHBOURHOOD_DECREASE);
			}
		}

		// enter the next input
		int num_pattern = (int) (rand.nextFloat() * TRAINING_SET.length);
		System.out.println("*** Using training pattern " + num_pattern + " ***");
		try {
			input.enterVector(TRAINING_SET[num_pattern]);
		}
		catch (CodingException e) {
			e.printStackTrace();
		}

		// remember the Tout
		next_Tout = getTime() + T;

		// and give the network enough time to settle
		timer.activateIn(T);

		currentLearningCycle++;
		if (currentLearningCycle % 100 == 0) {
			System.out.println("Current learning cycle is " + currentLearningCycle + ", weights are now:");
			// collect the weights
			double in_w[][] = new double[NUMBER_INPUTS+1][NUMBER_NEURONS];
			for (int j=0; j<NUMBER_NEURONS; j++) {
				Synapse[] syns = neurons[j].getInputSynapses();
				for (int i=0; i<NUMBER_INPUTS+1; i++) {
					in_w[i][j] = ((PLSynapse) syns[i]).getWeight();
					System.out.print(in_w[i][j]);
				}
				System.out.println();
			}

			try {
				FileOutputStream f = new FileOutputStream(FILENAME_COMPUTEDWEIGHTS);
				ObjectOutputStream s = new ObjectOutputStream(f);
				s.writeObject(in_w);
			}
			catch (Exception e2) {
				System.err.println("Aie! Fix this first!");
				System.err.println(e2.getMessage());
			}
		}
	}

	public void addSimEventComponent (mosaic.sim.kernel.simevent.SimEventComponent c) {
	}
	public void removeSimEventComponent (mosaic.sim.kernel.simevent.SimEventComponent c) {
	}
	public void addTimeEventComponent (mosaic.sim.kernel.simevent.SimEventComponent c) {
	}
	public void removeTimeEventComponent (mosaic.sim.kernel.simevent.SimEventComponent c) {
	}
}
