/*****************************************************************************
 *  Project: DEVSNeuron
 *  Description: discrete event-based simulation of spiking neural networks
 *  Filename: $RCSfile: ScrollingView.java,v $
 *  Author: Rene Mayrhofer
 *  Copyright: Rene Mayrhofer 2001
 *  Project version tag: $Name:  $
 *  File version: $Revision: 1.8 $
 *  Last changed at: $Date: 2002/05/16 20:43:35 $
 *  Last changed by: $Author: rene $
 *****************************************************************************/

package mosaic.sim.neuron.visualisation;

import java.awt.*;
import javax.swing.*;
import mosaic.sim.neuron.*;
import mosaic.sim.kernel.*;

/** This class implements a horizontally scrolling view for 2D trajectories. It
 *  can also synchronize with other ScrollingViews so that multiple views
 *  always show the same simulation timeframe.
 */
public class ScrollingView extends JPanel {

	/** The minimum time (horizontal) shown in the view. */
	protected double tmin = 0;
	/** The maximum time (horizontal) shown in the view. */
	protected double tmax = 10;
	/** The minimum vertical value shown in the view. */
	protected double ymin = 0;
	/** The maximum vertical value shown in the view. */
	protected double ymax = 10;

	/** The position on the time frame where to show the current simulation
	 *  time. This is from left to right.
	 */
	protected double currentTimePosition = 0.9;

	/** The ViewSynchronizer used to synchronize with other ScrollingViews. */
	protected ViewSynchronizer synchronizer = null;

	/** Transform world coordinates to screen graphics coordinates.
	 *  @param t The simulation time to transform.
	 *  @param y The vertical value to transform.
	 *  @returns A 2D Point which can be drawn relative to the panel view
	 *           position.
	 */
	synchronized protected Point worldToGraphics(double t, double y) {
		Point graphics = new Point();

		// get the border width
		Insets insets = getInsets();
		int currentWidth = getWidth() - insets.left - insets.right;
		int currentHeight = getHeight() - insets.top - insets.bottom;

		// the point insets.left / insets.top is the first that can be painted
		// on
		graphics.x = (int) ((double) (t - tmin) *
					  (double) currentWidth / (double) (tmax - tmin)) +
					  insets.left;
		graphics.y = currentHeight - insets.bottom -
					 (int) ((double) (y - ymin) *
					  (double) currentHeight / (double) (ymax - ymin));
		/*System.out.println("worldToGraphics: t=" + t + ", y=" + y +
			", tmin=" + tmin + ", tmax=" + tmax + ", x=" + graphics.x +
			", y=" + graphics.y);*/
		return graphics;
	}

	/** Creates a ScroolingView. */
	public ScrollingView() {
		// for performance reasons
		setOpaque(true);
	}

	/** Creates a ScrollingView and associates a ViewSynchronizer with it. */
	public ScrollingView(ViewSynchronizer sync) {
		super();
		synchronizer = sync;
	}

	/** Overrides the default implementation from JPanel to indicate the
	 *  current simulation time with a vertical line.
	 */
	synchronized public void paintComponent(Graphics g) {
		// paint the background
		super.paintComponent(g);

		// mark the current time position
		double currentTime = SimApplication.getCurrentSimApplication().getTime();
		Point from = worldToGraphics(currentTime, ymin);
		Point to = worldToGraphics(currentTime, ymax);
		g.setColor(Color.lightGray);
		g.drawLine(from.x, from.y, to.x, to.y);
	}

	/** This method scrolls the view horizontally, so that the current
	 *  simulation time is contained within the view. It does not scale the
	 *  view, but it scrolls it.
	 */
	synchronized public void scrollToSimulationTime(double currentTime) {
		double timeframe = tmax - tmin;
		/*System.out.println("ScrollingView.scrollToSimulationTime called with " +
			currentTime + ", tmin=" + tmin + ", tmax=" + tmax +
			", timeframe=" + timeframe);*/
		// shift tmin and tmax so that currentTime is displayed at the right
		// of the current viewport
		tmax = currentTime + (1 - currentTimePosition) * timeframe;
		tmin = tmax - timeframe;
		/*System.out.println("tmin=" + tmin + ", tmax=" + tmax);*/
		repaint();
	}

	/** Updates the current view (by calling scrollToSimulationTime with the
	 *  current simulation time) when no ViewSynchronizer is associated or
	 *  uses to associated ViewSynchronizer to scroll all synchronized views.
	 *  This method is preferrable over scrollToSimulationTime for direct use,
	 *  scrollToSimulationTime should only be used by ViewSynchronizer.
	 */
	synchronized public void updateView() {
		// synchronize other views if this one is located in a group
		double currentTime = SimApplication.getCurrentSimApplication().getTime();
		if (synchronizer != null)
			synchronizer.scrollToSimulationTime(currentTime);
		else
			scrollToSimulationTime(currentTime);
	}

	/** Returns the minimum simulation time shown.
	 *  @see #tmin
	 */
	synchronized public double getTMin() {
		return tmin;
	}

	/** Sets the minimum simulation time shown.
	 *  @see #tmin
	 */
	synchronized public void setTMin(double tmin) {
		/*System.out.println("ScrollingView.setTMin called with " + tmin);*/
		this.tmin = tmin;
	}

	/** Returns the maximum simulation time shown.
	 *  @see #tmax
	 */
	synchronized public double getTMax() {
		return tmax;
	}

	/** Sets the maximum simulation time shown.
	 *  @see #tmax
	 */
	synchronized public void setTMax(double tmax) {
		/*System.out.println("ScrollingView.setTMax called with " + tmax);*/
		this.tmax = tmax;
	}

	/** Returns the minimum vertical value shown.
	 *  @see #ymin
	 */
	synchronized public double getYMin() {
		return ymin;
	}

	/** Sets the minimum vertical value shown.
	 *  @see #ymin
	 */
	synchronized public void setYMin(double ymin) {
		this.ymin = ymin;
	}

	/** Returns the maximum vertical value shown.
	 *  @see #ymax
	 */
	synchronized public double getYMax() {
		return ymax;
	}

	/** Sets the maximum vertical value shown.
	 *  @see #ymax
	 */
	synchronized public void setYMax(double ymax) {
		this.ymax = ymax;
	}

	/** Sets the current time position.
	 *  @see #currentTimePosition
	 */
	synchronized public void setCurrentTimePosition(double currentTimePosition) {
		this.currentTimePosition = currentTimePosition;
	}

	/** Returns the current time position.
	 *  @see #currentTimePosition
	 */
	synchronized public double getCurrentTimePosition() {
		return currentTimePosition;
	}

	/** Returns the associated ViewSynchronizer.
	 *  @see #synchronizer
	 */
	public ViewSynchronizer getSynchronizer() {
		return synchronizer;
	}

	/** Sets the associated ViewSynchronizer.
	 *  @see #synchronizer
	 */
	public void setSynchronizer(ViewSynchronizer sync) {
		synchronizer = sync;
	}
}
