package mosaic.util;

/**
 * Queue
 * implements a simple queue internally using an array to hold the elements.
 * Growing of the queue is supported.
 * This class is not thread safe.
 *
 * time complexity:
 * enqueue / dequeue / getFirst / getLast / getElementAt(i): O(1)
 * increaseCapacity : O(N)
 *
 * @author Thomas Hilpold
 **/
public class Queue extends Object {

    private Object[] elementData;
    private int first;
    private int last;
    private int capacityIncrement;
    private int elementCount;

    /**
     * creates a Queue with an initialcapacity, which will be incremented by
     * capacityIncrement eah time the current capacity limit is reached.
     * @param capacityIncrement zero is allowed (non growing queue)
     */
    public Queue(int initialCapacity, int capacityIncrement) {
        elementData = new Object[initialCapacity];
        first = -1;
        last = -1;
        elementCount = 0;
        this.capacityIncrement = capacityIncrement;
    }

    /** gets the first element in the queue without dequeueing*/
    public Object getFirst() {
        return elementData[first];
    }

    /** gets the last element in the queue without dequeueing*/
    public Object getLast() {
        return elementData[last];
    }

    /** inserts an object to the beginning of the queue */
    public void enqueue(Object obj) {
        if (elementCount >= elementData.length) increaseCapacity();
        if (elementCount == 0) {
            first = 0;
            last = 0;
        } else {
            first = first - 1;
            if (first <= -1) first = elementData.length - 1;
        }
        elementCount++;
        elementData[first] = obj;
    }

    /** removes one object at the end of the queue */
    public Object dequeue() {
        Object ret;

        if (elementCount > 0) {
            ret = elementData[last];
            elementData[last] = null;
            last--;
            if (last == -1) last = elementData.length - 1;
            elementCount--;
            if (elementCount == 0) {
                last = -1;
                first = -1;
            }
        } else {

            /* queue empty */
            ret = null;
        }
        return ret;
    }

    /** shows at least all elements in the queue */
    public String toString() {
        String ret = "Queue[";

        for (int i = 0; i < elementCount; i++) {
            ret += getElementAt(i);
            if (i < elementCount - 1) ret += ";";
        }
        ret += "] \n count = " + elementCount + "\n capaciy = " + elementData.length;
        ret += "\n first = " + first + "\n last = " + last;
        /*
        ret += "\nelementData = ";
        for (int i = 0; i <elementData.length; i++) {
            ret += elementData[i] + ";";
        }
        ret += '\n';
        */
        return ret;
    }

    /**
     * returns the element at the given queue position
     * @param position 0..getQueueLength()-1; otherwise null will be returned
     */
    public Object getElementAt(int position) {
        if ((position < elementCount) && (position >= 0)) {
            position += first;
            position = position % elementData.length;
            return elementData[position];
        } else {
            return null;
        }
    }

    public boolean isEmpty() {
        return (elementCount == 0);
    }

    /** the queue can only be full if the capacityIncrement is set to zero */
    public boolean isFull() {
        return ((elementCount == elementData.length) &&
                    (capacityIncrement == 0));
    }

    /** returns the length of the queue */
    public int getQueueLength() {
        return elementCount;
    }

    private void increaseCapacity() {
        /* arraycopy used */
        Object[] newData;

        newData = new Object[elementData.length + capacityIncrement];
        if (elementCount > 0) System.arraycopy(elementData, 0, newData, 0, last + 1);
        if (first > last) {
            for (int i = first; i < elementData.length; i++) {
                newData[i + capacityIncrement] = elementData[i];
            }
            first = first + capacityIncrement;
        }
        elementData = newData;
    }

    /** performs a test */
/*    public static void main(String[] argv) {
        Queue q = new Queue(1000000, 0);
        System.out.println(" Assert 0 elems ");
        System.out.println(q.toString());
        q.enqueue("1");
        System.out.println(q.toString());
        q.enqueue("2");
        System.out.println(q.toString());
        q.enqueue("3");
        System.out.println(q.toString());
        q.enqueue("4");
        System.out.println(" Assert 4 elems ");
        System.out.println(q.toString());
        q.dequeue();
        System.out.println(q.toString());
        q.dequeue();
        System.out.println(q.toString());
        q.dequeue();
        System.out.println(q.toString());
        q.dequeue();
        System.out.println(" Assert 0 elem ");
        System.out.println(q.toString());
        q.dequeue();
        System.out.println(q.toString());
        tools.StopWatch.start();
        for (int i = 0; i < 1000000; i++) {
            q.enqueue(new Integer(i));
        }
        for (int i = 0; i < 1000000; i++) {
            q.dequeue();
        }
        double time = tools.StopWatch.stop(" 1000000 enq + deq ");
        System.out.println("q+e per ms = " + 100000.0 / time / 1000.0);
        System.out.println(q.toString());
        try {
            Thread.sleep(1000000);
        } catch (InterruptedException e) {
        }
    }
*/
}

