package mosaic.util.custom;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.lang.reflect.Array;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import javax.swing.border.*;

/**
 * Class, whose Instances can be used as a visual CustomEditor for Arrays.
 * It is allowed that the CustomArrayEditorAdapter contains a primitive array.
 *
 * changes history:
 * 15.11.1999   Hilpold: last change in array not propagated
 *                       -> added FocusListener to arrTable
 *
 * @version 17 Feb 1999
 * @author Thomas Hilpold
 */
public class CustomArrayPropEditor extends JPanel
    implements ActionListener, TableModelListener, ListSelectionListener,
        FocusListener {

    //
    // Data
    //

    protected JButton addButton;
    protected JButton delButton;
    protected JTable arrTable;
    protected CustomArrayTableModel arrTableModel;
    protected ListSelectionModel arrSelectionModel;
    protected JScrollPane arrScrollPane;
    protected CustomArrayEditorAdapter pe;

    //PrimitiveTypes are wrapped into arrays of their conterpart object
    protected boolean arrayIsPrimitive;

    //
    // Debugging output
    //
    private boolean[] DEBUG = new boolean[] {
        false,  // trace primitive Arrays and tablechanged Events
        true   // trace table focus management
        };


    //
    // Construction
    //

    /**
     * Asserts that ArrayEditor contains not null
     * will accept built primitive type arrays.
     */
    public CustomArrayPropEditor(CustomArrayEditorAdapter aPropertyEditor) {
        pe = aPropertyEditor;
        Object value = pe.getAsArray();
        Object[] arrValue;
        //assert value not null
        arrayIsPrimitive = value.getClass().getComponentType().isPrimitive();
        if (arrayIsPrimitive) {
            if (DEBUG[0]) OutFrame.out("PRIMITIVE !");
            arrValue = wrapPrimitiveArray(value);
        }
        else {
            arrValue = (Object[]) value;
        }
        // postcondition: value components are not primitive Types
        //OutFrame.out(this.toString() + " NEW, PE = " + aPropertyEditor +" ****");
        arrTableModel = new CustomArrayTableModel();
        arrTableModel.setData(arrValue);
        //setPreferredSize(new Dimension(200, 200));
        setLayout(new BorderLayout(5, 5));
        JPanel p = new JPanel();
        p.setMinimumSize(new Dimension(200,20));
        p.setLayout(new BorderLayout(10, 5));
        addButton = new JButton("Add Index");
        p.add(addButton, "West");
        delButton = new JButton("Delete Index");
        delButton.setEnabled(false);
        p.add(delButton, "East");
        add(p, "South");
        arrTable = new JTable(arrTableModel);
        //register to listen
        addButton.addActionListener(this);
        delButton.addActionListener(this);
        arrTableModel.addTableModelListener(this);
        //Detecting user selections
        arrSelectionModel = arrTable.getSelectionModel();
        arrSelectionModel.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        arrSelectionModel.addListSelectionListener(this);
        //setBgColor
        arrTable.setBackground(this.getBackground());
        //set Editor and Renderer
        arrTable.getColumnModel().getColumn(0).setMaxWidth(60);
        arrTable.getColumnModel().getColumn(0).setMinWidth(40);
        arrTable.getColumnModel().getColumn(0).setWidth(50);
        arrTable.getColumnModel().getColumn(1).setMinWidth(50);
        arrTable.getColumnModel().getColumn(1).setCellEditor(
            new CustomArrayCellEditor(arrValue.getClass().getComponentType())
            );

        arrTable.getColumnModel().getColumn(1).setCellRenderer(
            new DefaultTableCellRenderer()
            );
        arrTable.setAutoResizeMode(JTable.AUTO_RESIZE_LAST_COLUMN);
        arrTable.addFocusListener(this);
        arrScrollPane = new JScrollPane(arrTable);
        arrScrollPane.setMinimumSize(new Dimension(200, 200));
        add (arrScrollPane,"Center");
    }

    //
    // TableModelListener Interface
    //

    /** Updates the PropertyEditors value
     *  must be called each time the table changes
     *  to keep our PropertyEditor up to date
     */
    public void tableChanged(TableModelEvent e) {
        Object value;
        if (DEBUG[0]) OutFrame.out("TableChanged called");
        if (e.getSource() == arrTableModel) {
            value = arrTableModel.getData();
            // check if the PropertyEditor is for primitive Types
            if(arrayIsPrimitive) {
                value = unwrapObjectArray((Object[])value);
            }
            if (DEBUG[0]) OutFrame.out("Try to set Array"+ value);
            pe.setAsArray(value);
        }
    }

    //
    // ActionListener Interface
    //

    /** Takes Actions when user adds or dels a line
     *  currently only the last line can be deleted / after the last inserted
     * this is subject to change PENDING
     */
    public void actionPerformed(ActionEvent e) {
        if(e.getSource() == addButton) {
            arrTableModel.addIndex();
        }
        else if (e.getSource() == delButton) {
            int index = arrSelectionModel.getMinSelectionIndex();
            if (index >= 0) {
                arrTableModel.delIndex(index);
                if (index == 0) {
                    arrSelectionModel.clearSelection();
                    delButton.setEnabled(false);
                }
                else {
                    arrSelectionModel.setSelectionInterval(index - 1, index - 1);
                    delButton.setEnabled(true);
                }
            }
        }
    }

    //
    // ListSelectionListener Method
    //
    public void valueChanged(ListSelectionEvent e) {
        delButton.setEnabled(!arrSelectionModel.isSelectionEmpty());
    }




    //
    // Methods for wrapping and unwrapping primitives
    //

    private Object[] wrapPrimitiveArray(Object arr) {
        Object[] ret;
        int length;
        Class cType;
        Class wrapperType;

        cType = arr.getClass().getComponentType();
        length = Array.getLength(arr);
        // check for TYPE
        if (cType.equals(Boolean.TYPE  )) wrapperType = Boolean.class;
        else if (cType.equals(Character.TYPE)) wrapperType = Character.class;
        else if (cType.equals(Byte.TYPE     )) wrapperType = Byte.class;
        else if (cType.equals(Double.TYPE   )) wrapperType = Double.class;
        else if (cType.equals(Float.TYPE    )) wrapperType = Float.class;
        else if (cType.equals(Integer.TYPE  )) wrapperType = Integer.class;
        else if (cType.equals(Long.TYPE     )) wrapperType = Long.class;
        else if (cType.equals(Short.TYPE    )) wrapperType = Short.class;
        else {
            OutFrame.out(this.toString() + " COULD NOT WRAP TYPE ");
            wrapperType = null;
        }
        // create new OBJECT[]
        ret = (Object[]) Array.newInstance(wrapperType, length);
        // fill OBJECT[]
        for (int i=0; i < length; i++) {
            // Arrays get automatically wrapps primitive Types
            ret[i] = Array.get(arr, i);
        }
    return ret;
    }

    private Object unwrapObjectArray(Object[] arr) {
        Object ret;
        Class cType = arr.getClass().getComponentType();
        Class primitiveType;
        if (cType.equals(Boolean.class  )) primitiveType = Boolean.TYPE;
        else if (cType.equals(Character.class)) primitiveType = Character.TYPE;
        else if (cType.equals(Byte.class     )) primitiveType = Byte.TYPE;
        else if (cType.equals(Double.class   )) primitiveType = Double.TYPE;
        else if (cType.equals(Float.class    )) primitiveType = Float.TYPE;
        else if (cType.equals(Integer.class  )) primitiveType = Integer.TYPE;
        else if (cType.equals(Long.class     )) primitiveType = Long.TYPE;
        else if (cType.equals(Short.class    )) primitiveType = Short.TYPE;
        else {
            OutFrame.out(this.toString() + " COULD NOT UNWRAP TYPE ");
            primitiveType = null;
        }
        ret = Array.newInstance(primitiveType, arr.length);
        for (int i=0; i < arr.length; i++) {
            // automatically unwrapps Objects to primitives
            // we can not set a field of primitive Types null !!
            if (arr[i] != null) Array.set(ret, i, arr[i]);
        }
    return ret;
    }

    //
    // FocusListener Interface
    //

    public void focusGained(FocusEvent e){};

    /**
     * if the table is currently editing a cell, the celleditor will be
     * urged to stop cell editing
     */
    public void focusLost(FocusEvent e){
        if (DEBUG[1])System.err.println("TABLE FOCUS LOST ");
        if (arrTable.isEditing()) {
            if (DEBUG[1])System.err.println("TABLE EDITING ");
            arrTable.getCellEditor(0, 1).stopCellEditing();
        } else {
            if (DEBUG[1])System.err.println("TABLE NOT EDITING ");
        }
    };
}
