package mosaic.vis.output;

import java.awt.*;
import javax.swing.*;
import mosaic.sim.kernel.variable.*;
import mosaic.sim.kernel.control.AnimatedView;
import mosaic.vis.general.VariableView;


abstract public class WheelView extends VariableView implements AnimatedView {

    public static final int POINT_MODE = 0;
    public static final int LINE_MODE  = 1;

    protected static final double MIN_ANGLE_INCREMENT = 0.3;

    //
    // Construction
    //

    /** point mode */
    public WheelView() {
        this(3, 4, POINT_MODE);
    }

    /** point mode */
    public WheelView(int nrOfSpokes) {
        this(nrOfSpokes, 4, POINT_MODE);
    }


    /**
     * defaults to point mode
     * @param spokeWidth width of the spokeLine or size of the points
     */
    public WheelView(int nrOfSpokes, int spokeWidth, int mode) {
        lastPaintTime = 0.0;
        setNrOfSpokes(nrOfSpokes);
        curAngle = 0.0f;
        isOn = true;
        setSpokeWidth(spokeWidth);
        setMode(mode);
        setCircleWidth(2);
        curColor = onColor;
    }

    //
    // Properties
    //

    // Property NrOfSpokes
    protected int nrOfSpokes = 4;

     /** Get the number of the spokes */
    public int getNrOfSpokes() {
        return nrOfSpokes;
    }

    /** Set the number of the spokes */
    public void setNrOfSpokes(int  nr) {
        if (nr >= 1) {
            this.nrOfSpokes = nr;
            nrOfSpokesChanged();
        }
    }

    private void nrOfSpokesChanged() {
        pointsX = new int[nrOfSpokes];
        pointsY = new int[nrOfSpokes];
        angleBetweenSpokes = 360.0f/nrOfSpokes;
        calcPoints();
        repaint();
    }

    // Property spokeWidth
    protected int spokeWidth = 2;

     /** Get the number of the spokes */
    public int getSpokeWidth() {
        return spokeWidth;
    }

    /** Set the number of the spokes */
    public void setSpokeWidth(int w) {
        if (w > 0) {
            spokeWidth = w;
            if (mode == LINE_MODE) {
                stroke = new BasicStroke(w);
            }
            repaint();
        }
    }

    // Property circleWidth
    protected int circleWidth;

     /** Get the number of the spokes */
    public int getCircleWidth() {
        return circleWidth;
    }

    /** Set the number of the spokes */
    public void setCircleWidth(int w) {
        if (w >= 0) {
            circleWidth = w;
            circleStroke = new BasicStroke(w);
            repaint();
        }
    }

    // Property OnColor
    Color onColor = Color.darkGray;

    public void setOnColor( Color c) {
        if (c != null) {
            onColor = c;
            repaint();
        }
    }

    public Color getOnColor () {
        return onColor;
    }

    // Property offColor
    Color offColor = Color.lightGray;

    public void setOffColor( Color c) {
        if (c != null) {
            offColor = c;
            repaint();
        }
    }

    public Color getOffColor () {
        return offColor;
    }

    // Property mode
    protected int mode = LINE_MODE;

    public void setMode( int mode) {
        if (mode == POINT_MODE || mode == LINE_MODE) {
           this.mode = mode;
        } else {
            System.err.println(this + " illegal mode");
            this.mode = POINT_MODE;
        }
        setSpokeWidth(getSpokeWidth());
        repaint();
    }

    public int getMode () {
        return mode;
    }

    //
    // Data
    //

    private  int[] pointsX;  //array of Points for the next frame
    private  int[] pointsY;
    private  int radius;     //the radius of the wheel
    private  int center;     // centerX == centerY center of the circle
    private  double angleBetweenSpokes;
    private  double curAngle;
    protected  double curAngleIncrement = 1.0f; //just a dummy value
    protected  boolean isOn;
    private  Stroke stroke;      //for spokes if LINE_MODE
    private  Stroke circleStroke; //for the circle
    private  double lastPaintTime;
    private  Color curColor;


    /** will set the spokes to inital position */
    public void reset() {
        curAngle = 0;
        calcPoints();
        curColor = onColor;
        repaint();
    }

    private synchronized void calcPoints() {
        center = (int)(getWidth() / 2.0);
        radius = (int)((getWidth() / 2.0) - spokeWidth - 1);
        for (int i = 0; i < nrOfSpokes; i++) {
            pointsX[i] = center + (int)(radius * Math.sin((angleBetweenSpokes * i + curAngle) / 360.0 * 2 * Math.PI));
            pointsY[i] = center + (int)(radius * Math.cos((angleBetweenSpokes * i + curAngle) / 360.0 * 2 * Math.PI));
        }
    }

    /**
     * Will move the spokes to the next frame
     * curAngleIncrement should be set properly.
     * Parameter time ignored here, used by subclasses
     */
    public boolean performAnimationStep(double time) {
        boolean doPaint = ((curAngleIncrement != 0.0) && (isOn));
        if (doPaint) {
            double actIncrement;
            if (curAngleIncrement > 0.0) {
                actIncrement = Math.max(curAngleIncrement*(time-lastPaintTime), MIN_ANGLE_INCREMENT);
            } else {
                actIncrement = Math.min(curAngleIncrement*(time-lastPaintTime), -MIN_ANGLE_INCREMENT);
            }
            curAngle = (curAngle + actIncrement) % 360.0f;
            // paint old spokes with offColor
            curColor = offColor;
            paintCached();

            calcPoints();
            // paint new spokes with onColor
            curColor = onColor;
            paintCached();
        }
        lastPaintTime = time;
        return doPaint;
    }

    //boolean paintCallback = false;
    public void paint(Graphics g) {
        paintBackgroundWithGfx(g);
        calcPoints();
        if (isOn) {
            curColor = onColor;
            paintWithGfx(g);
        }
    }

    final public void paintBackgroundWithGfx(Graphics g) {
        if (isShowing()) {
            Graphics2D g2 = (Graphics2D) g;
            // draw background
            g2.setColor(getOffColor());
            g2.fillOval(0 + spokeWidth/2, 0 + spokeWidth/2, getWidth() - spokeWidth/2, getWidth() - spokeWidth/2);
            // draw wheel
            if (circleWidth > 0) {
                Stroke oldStroke = g2.getStroke();
                g2.setColor(getForeground());
                g2.setStroke(circleStroke);
                g2.drawOval(0, 0, getWidth() - 1, getWidth() - 1);
                g2.setStroke(oldStroke);
            }
        }
    }

    final public void paintWithGfx(Graphics g) {
        Graphics2D g2 = (Graphics2D) g;
        if (isShowing()) {
            g2.setColor(curColor);
            if (mode == LINE_MODE) {
                int nrLines = nrOfSpokes / 2;
                for (int i = 0; i < nrLines; i++) {
                    g2.setStroke(stroke);
                    g2.drawLine(pointsX[i], pointsY[i], pointsX[i+nrLines], pointsY[i+nrLines]);
                }
            } else {
                // POINT_MODE
                for (int i = 0; i < nrOfSpokes; i++) {
                    g2.fillOval(pointsX[i] - spokeWidth/2, pointsY[i] - spokeWidth/2, spokeWidth, spokeWidth);
                }
            }
        }
    }
}
